-- Contents:
--    Tween.lua
--    GlobalWrappers.lua
--    Thread.lua
--    Semaphore.lua
--    Async.lua
--    ThreadCompat.lua
--    Utils.lua
--    Debugger.lua
--    AutoDoc.lua
--    DocViewer.lua

------------------------------
-- Tween.lua
------------------------------
do
-- tween.lua
-- v1.0.1 (2012-02)

-- Based on:
-- Enrique García Cota - enrique.garcia.cota [AT] gmail [DOT] com
-- tweening functions for lua
-- inspired by jquery's animate function

-- Modified for Codea by John Millard 2013-02
-- Modified for Codea+ by Oliver Steptoe 2022-04
-----------------------------------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------

tween = {}

-- private stuff

local tweens = {}

local function isCallable(f)
  local tf = type(f)
  if tf == 'function' then return true end
  if tf == 'table' then
    local mt = getmetatable(f)
    return (type(mt) == 'table' and type(mt.__call) == 'function')
  end
  return false
end

local function copyTables(destination, keysTable, valuesTable)
  valuesTable = valuesTable or keysTable
  for k,v in pairs(keysTable) do
    if type(v) == 'table' then
      destination[k] = copyTables({}, v, valuesTable[k])
    else
      destination[k] = valuesTable[k]
    end
  end
  return destination
end

local function checkSubjectAndTargetRecursively(subject, target, path)
  path = path or {}
  local targetType, newPath
  for k,targetValue in pairs(target) do
    targetType, newPath = type(targetValue), copyTables({}, path)
    table.insert(newPath, tostring(k))
    if targetType == 'number' then
      assert(type(subject[k]) == 'number', "Parameter '" .. table.concat(newPath,'/') .. "' is missing from subject or isn't a number")
    elseif targetType == 'userdata' then
      assert(type(subject[k]) == 'userdata', "Parameter '" .. table.concat(newPath,'/') .. "' is missing from subject or isn't a userdata")
    elseif targetType == 'table' then
      checkSubjectAndTargetRecursively(subject[k], targetValue, newPath)
    else
      assert(targetType == 'number', "Parameter '" .. table.concat(newPath,'/') .. "' must be a number or table of numbers")
    end
  end
end

local function getEasingFunction(easing)
  easing = easing or "linear"

  if type(easing) == 'table' then
    easing = easing['easing'] or "linear"
    end
    
  if type(easing) == 'string' then
    local name = easing
    easing = tween.easing[name]
    assert(type(easing) == 'function', "The easing function name '" .. name .. "' is invalid")
  end
  return easing
end

local function getloopFunction(loop)
  loop = loop or "once"
  
  if type(loop) == 'table' then
    loop = loop['loop'] or "once"
    
    if type(loop) == 'string' then
      local name = loop
      loop = tween.loop[name]
      assert(type(loop) == 'function', "The loop function name '" .. name .. "' is invalid")
    end
  else
    loop = tween.loop.once
  end

  return loop
end

local function checkStartParams(time, subject, target, options, callback)
  local easing = getEasingFunction(options)
  local loop = getloopFunction(options)

  assert(type(time) == 'number' and time > 0, "time must be a positive number. Was " .. tostring(time))
  local tsubject = type(subject)
  assert(tsubject == 'table' or tsubject == 'userdata', "subject must be a table or userdata. Was " .. tostring(subject))
  assert(type(target)== 'table', "target must be a table. Was " .. tostring(target))
  assert(isCallable(easing), "easing must be a function or functable. Was " .. tostring(easing))
  assert(isCallable(loop), "loop must be a function or functable. Was " .. tostring(loop))
  assert(callback==nil or isCallable(callback), "callback must be nil, a function or functable. Was " .. tostring(time))
  checkSubjectAndTargetRecursively(subject, target)

end

local function newTween(time, subject, target, options, callback, args)
  local self = {
    time = time,
    subject = subject,
    target = target,
    easing = getEasingFunction(options),
    loop = getloopFunction(options),
    callback = callback,
    args = args,
    initial = copyTables({}, target, subject),
    running = 0
  }
  tweens[self] = self
  return self
end

local function easeWithTween(self, subject, target, initial)
  local t,b,c,d

  for k,v in pairs(target) do
    if type(v)=='table' then
      easeWithTween(self, subject[k], v, initial[k])
    else
      t,b,c,d = self.running, initial[k], v - initial[k], self.time
      
      if self.loop then
        t = self.loop(t,d)
      end

      subject[k] = self.easing(t,b,c,d)
    end
  end
end

local function updateTween(self, dt)
    if Debugger.thread and self.ignoreDebug ~= true then return end
    self.running = self.running + dt
    easeWithTween(self, self.subject, self.target, self.initial)
end

local function hasExpiredTween(self)
  if self.loop == tween.loop.once then
    return self.running >= self.time
  else
    return false
  end
end

local function finishTween(self)
  self.overshoot = self.running - self.time
  self.running = self.overshoot

  tween.stop(self)

    if self.callback then
        if self.noAsyncCallback then
            self.callback(table.unpack(self.args))
        else
            Thread.runAsync(self.callback, table.unpack(self.args))
        end
    end

  if self.next then
    self.next.initial = copyTables(self.next.initial, self.target, self.subject)
    tweens[self.next] = self.next
  end

end

local function resetTween(self)
  self.running = 0
  easeWithTween(self, self.subject, self.target, self.initial)
  --copyTables(self.subject, self.initial)
end

-- paths

local function pathFactory(subject, values)
    local len = #values
    
    assert(len >= 1, "Path length must be one or greater")
    
    local keys = {}
    for k,v in pairs(values[1]) do
        table.insert(keys,k)
    end
    
    local pathFunc = function(t,k)
        
        t = (1 + (t*(len-1)))
        
        local i2 = math.floor(t)
        local i1 = math.max(i2 - 1, 1)
        local i3 = math.min(i2 + 1, len)
        local i4 = math.min(i3 + 2, len)
        
        t = t - i2
        
        local p1 = values[i1][k]
        local p2 = values[i2][k]
        local p3 = values[i3][k]
        local p4 = values[i4][k]
        
        local t2 = t*t
        local t3 = t*t*t
        
        return 0.5 * ( (2*p2) + 
                       (-p1+p3)*t + 
                       (2*p1 - 5*p2 + 4*p3 - p4)*t2 + 
                       (-p1 + 3*p2 - 3*p3 + p4)*t3)
    end
    
    local proxy = {}
    local mt = 
    {
        __index = function(table,k)
            if k == 't' then
                return 0
            end
            return rawget(table,k)
        end,
        
        __newindex = function(table,k,v)
            if k == 't' then
                for i,key in pairs(keys) do
                    subject[key] = pathFunc(v,key)
                end
            else
                rawset(table,k,v)
            end
        end 
    }
    setmetatable(proxy,mt)
    return proxy
end



-- easing

-- Adapted from https://github.com/EmmanuelOga/easing. See LICENSE.txt for credits.
-- For all easing functions:
-- t = time == how much time has to pass for the tweening to complete
-- b = begin == starting property value
-- c = change == ending - beginning
-- d = duration == running time. How much time has passed *right now*

local pow = function(x, y) return x^y end
local sin, cos, pi, sqrt, abs, asin = math.sin, math.cos, math.pi, math.sqrt, math.abs, math.asin

-- linear
local function linear(t, b, c, d) return c * t / d + b end

-- quad
local function inQuad(t, b, c, d) return c * pow(t / d, 2) + b end
local function outQuad(t, b, c, d)
  t = t / d
  return -c * t * (t - 2) + b
end
local function inOutQuad(t, b, c, d)
  t = t / d * 2
  if t < 1 then return c / 2 * pow(t, 2) + b end
  return -c / 2 * ((t - 1) * (t - 3) - 1) + b
end
local function outInQuad(t, b, c, d)
  if t < d / 2 then return outQuad(t * 2, b, c / 2, d) end
  return inQuad((t * 2) - d, b + c / 2, c / 2, d)
end

-- cubic
local function inCubic (t, b, c, d) return c * pow(t / d, 3) + b end
local function outCubic(t, b, c, d) return c * (pow(t / d - 1, 3) + 1) + b end
local function inOutCubic(t, b, c, d)
  t = t / d * 2
  if t < 1 then return c / 2 * t * t * t + b end
  t = t - 2
  return c / 2 * (t * t * t + 2) + b
end
local function outInCubic(t, b, c, d)
  if t < d / 2 then return outCubic(t * 2, b, c / 2, d) end
  return inCubic((t * 2) - d, b + c / 2, c / 2, d)
end

-- quart
local function inQuart(t, b, c, d) return c * pow(t / d, 4) + b end
local function outQuart(t, b, c, d) return -c * (pow(t / d - 1, 4) - 1) + b end
local function inOutQuart(t, b, c, d)
  t = t / d * 2
  if t < 1 then return c / 2 * pow(t, 4) + b end
  return -c / 2 * (pow(t - 2, 4) - 2) + b
end
local function outInQuart(t, b, c, d)
  if t < d / 2 then return outQuart(t * 2, b, c / 2, d) end
  return inQuart((t * 2) - d, b + c / 2, c / 2, d)
end

-- quint
local function inQuint(t, b, c, d) return c * pow(t / d, 5) + b end
local function outQuint(t, b, c, d) return c * (pow(t / d - 1, 5) + 1) + b end
local function inOutQuint(t, b, c, d)
  t = t / d * 2
  if t < 1 then return c / 2 * pow(t, 5) + b end
  return c / 2 * (pow(t - 2, 5) + 2) + b
end
local function outInQuint(t, b, c, d)
  if t < d / 2 then return outQuint(t * 2, b, c / 2, d) end
  return inQuint((t * 2) - d, b + c / 2, c / 2, d)
end

-- sine
local function inSine(t, b, c, d) return -c * cos(t / d * (pi / 2)) + c + b end
local function outSine(t, b, c, d) return c * sin(t / d * (pi / 2)) + b end
local function inOutSine(t, b, c, d) return -c / 2 * (cos(pi * t / d) - 1) + b end
local function outInSine(t, b, c, d)
  if t < d / 2 then return outSine(t * 2, b, c / 2, d) end
  return inSine((t * 2) -d, b + c / 2, c / 2, d)
end

-- expo
local function inExpo(t, b, c, d)
  if t == 0 then return b end
  return c * pow(2, 10 * (t / d - 1)) + b - c * 0.001
end
local function outExpo(t, b, c, d)
  if t == d then return b + c end
  return c * 1.001 * (-pow(2, -10 * t / d) + 1) + b
end
local function inOutExpo(t, b, c, d)
  if t == 0 then return b end
  if t == d then return b + c end
  t = t / d * 2
  if t < 1 then return c / 2 * pow(2, 10 * (t - 1)) + b - c * 0.0005 end
  return c / 2 * 1.0005 * (-pow(2, -10 * (t - 1)) + 2) + b
end
local function outInExpo(t, b, c, d)
  if t < d / 2 then return outExpo(t * 2, b, c / 2, d) end
  return inExpo((t * 2) - d, b + c / 2, c / 2, d)
end

-- circ
local function inCirc(t, b, c, d) return(-c * (sqrt(1 - pow(t / d, 2)) - 1) + b) end
local function outCirc(t, b, c, d)  return(c * sqrt(1 - pow(t / d - 1, 2)) + b) end
local function inOutCirc(t, b, c, d)
  t = t / d * 2
  if t < 1 then return -c / 2 * (sqrt(1 - t * t) - 1) + b end
  t = t - 2
  return c / 2 * (sqrt(1 - t * t) + 1) + b
end
local function outInCirc(t, b, c, d)
  if t < d / 2 then return outCirc(t * 2, b, c / 2, d) end
  return inCirc((t * 2) - d, b + c / 2, c / 2, d)
end

-- elastic
local function calculatePAS(p,a,c,d)
  p, a = p or d * 0.3, a or 0
  if a < abs(c) then return p, c, p / 4 end -- p, a, s
  return p, a, p / (2 * pi) * asin(c/a) -- p,a,s
end
local function inElastic(t, b, c, d, a, p)
  local s
  if t == 0 then return b end
  t = t / d
  if t == 1  then return b + c end
  p,a,s = calculatePAS(p,a,c,d)
  t = t - 1
  return -(a * pow(2, 10 * t) * sin((t * d - s) * (2 * pi) / p)) + b
end
local function outElastic(t, b, c, d, a, p)
  local s
  if t == 0 then return b end
  t = t / d
  if t == 1 then return b + c end
  p,a,s = calculatePAS(p,a,c,d)
  return a * pow(2, -10 * t) * sin((t * d - s) * (2 * pi) / p) + c + b
end
local function inOutElastic(t, b, c, d, a, p)
  local s
  if t == 0 then return b end
  t = t / d * 2
  if t == 2 then return b + c end
  p,a,s = calculatePAS(p,a,c,d)
  t = t - 1
  if t < 0 then return -0.5 * (a * pow(2, 10 * t) * sin((t * d - s) * (2 * pi) / p)) + b end
  return a * pow(2, -10 * t) * sin((t * d - s) * (2 * pi) / p ) * 0.5 + c + b
end
local function outInElastic(t, b, c, d, a, p)
  if t < d / 2 then return outElastic(t * 2, b, c / 2, d, a, p) end
  return inElastic((t * 2) - d, b + c / 2, c / 2, d, a, p)
end

-- back
local function inBack(t, b, c, d, s)
  s = s or 1.70158
  t = t / d
  return c * t * t * ((s + 1) * t - s) + b
end
local function outBack(t, b, c, d, s)
  s = s or 1.70158
  t = t / d - 1
  return c * (t * t * ((s + 1) * t + s) + 1) + b
end
local function inOutBack(t, b, c, d, s)
  s = (s or 1.70158) * 1.525
  t = t / d * 2
  if t < 1 then return c / 2 * (t * t * ((s + 1) * t - s)) + b end
  t = t - 2
  return c / 2 * (t * t * ((s + 1) * t + s) + 2) + b
end
local function outInBack(t, b, c, d, s)
  if t < d / 2 then return outBack(t * 2, b, c / 2, d, s) end
  return inBack((t * 2) - d, b + c / 2, c / 2, d, s)
end

-- bounce
local function outBounce(t, b, c, d)
  t = t / d
  if t < 1 / 2.75 then return c * (7.5625 * t * t) + b end
  if t < 2 / 2.75 then
    t = t - (1.5 / 2.75)
    return c * (7.5625 * t * t + 0.75) + b
  elseif t < 2.5 / 2.75 then
    t = t - (2.25 / 2.75)
    return c * (7.5625 * t * t + 0.9375) + b
  end
  t = t - (2.625 / 2.75)
  return c * (7.5625 * t * t + 0.984375) + b
end
local function inBounce(t, b, c, d) return c - outBounce(d - t, 0, c, d) + b end
local function inOutBounce(t, b, c, d)
  if t < d / 2 then return inBounce(t * 2, 0, c, d) * 0.5 + b end
  return outBounce(t * 2 - d, 0, c, d) * 0.5 + c * .5 + b
end
local function outInBounce(t, b, c, d)
  if t < d / 2 then return outBounce(t * 2, b, c / 2, d) end
  return inBounce((t * 2) - d, b + c / 2, c / 2, d)
end

tween.easing = {
  linear = linear,
  quadIn    = inQuad,    quadOut    = outQuad,    quadInOut    = inOutQuad,    quadOutIn    = outInQuad,
  cubicIn   = inCubic,   cubicOut   = outCubic,   cubicInOut   = inOutCubic,   cubicOutIn   = outInCubic,
  quartIn   = inQuart,   quartOut   = outQuart,   quartInOut   = inOutQuart,   quartOutIn   = outInQuart,
  quintIn   = inQuint,   quintOut   = outQuint,   quintInOut   = inOutQuint,   quintOutIn   = outInQuint,
  sineIn    = inSine,    sineOut    = outSine,    sineInOut    = inOutSine,    sineOutIn    = outInSine,
  expoIn    = inExpo,    expoOut    = outExpo,    expoInOut    = inOutExpo,    expoOutIn    = outInExpo,
  circIn    = inCirc,    circOut    = outCirc,    circInOut    = inOutCirc,    circOutIn    = outInCirc,
  elasticIn = inElastic, elasticOut = outElastic, elasticInOut = inOutElastic, elasticOutIn = outInElastic,
  backIn    = inBack,    backOut    = outBack,    backInOut    = inOutBack,    backOutIn    = outInBack,
  bounceIn  = inBounce,  bounceOut  = outBounce,  bounceInOut  = inOutBounce,  bounceOutIn  = outInBounce,
}

-- ping pong

local function round(num)
  if num >= 0 then return math.floor(num+.5)
  else return math.ceil(num-.5) end
end

local function triangleWave(x)
  return 1 - 2 * math.abs(round(0.5 * x) - 0.5*x)
end

local function frac(x)
  local a,b = math.modf(x)
  return b
end

local function sawtooth(x)
  return frac(0.5 * x)
end

local function forever(t,d)
  return sawtooth(t/(d/2)) * d
end

local function pingpong(t,d)
    return (1-triangleWave(t/d)) * d
end

local function once(t,d)
    return math.min(t,d)
end

tween.loop = 
{
    once = once,
    forever = forever,
    pingpong = pingpong
}

-- public functions

function tween.start(time, subject, target, options, callback, ...)
  checkStartParams(time, subject, target, options, callback)
  return newTween(time, subject, target, options, callback, {...})
end

function tween.sequence(...)
    local n = select('#', ...)
    local arg = {...}
    assert(n > 0, "a sequence must consist of at least one tween")

    local head = arg[1]
    local tail = arg[n]

    for i,id in pairs(arg) do
        local tw = tweens[id]
        
        tw.head = head
        tw.tail = tail

        if i < n then
            tw.next = arg[i+1]
        end

        -- remove all except the first tween
        if i > 1 and tw then
            tweens[id] = nil
        end
    end

    return arg[1]
end

function tween.path(time, subject, target, options, callback, ...)
  -- need to check that target is an array and contains valid keys
  return newTween(time, pathFactory(subject, target) , {t=1}, options, callback, {...})
end

function tween.delay(duration, callback, ...)
    local t = {}
    return tween(duration, t, {}, tween.easing.linear, callback, ...)
end

setmetatable(tween, { __call = function(t, ...) return tween.start(...) end })

function tween.play(id)
    tweens[id]=id
end

function tween.reset(id)
  local tw = tweens[id]
  if tw then
    resetTween(tw)
    tween.stop(tw)
  end
end

function tween.resetAll(id)
  for _,tw in pairs(tweens) do 
    resetTween(tw)
    --copyTables(tw.subject, tw.initial)
  end
  tween.stopAll()
end

function tween.stop(id)
  -- this tween object is the head of a sequence, stop all connected tweens
  if id.head and id.head == id then
    local next = id.next
    while next ~= nil do
      tweens[next]=nil
      next = next.next
    end
  end

  if id~=nil then 
    tweens[id]=nil
  end
end

function tween.stopAll()
  tweens = {}
end

function tween.hasExpired(id)
  local tw = tweens[id]
  if tw then
    return hasExpiredTween(id)
  end
  return true
end

function tween.update(dt)
    if dt <= 0 then
        return
    end
    
    local expired = {}
    for _,t in pairs(tweens) do
        updateTween(t, dt)
        if hasExpiredTween(t) then
            table.insert(expired, t)
        end
    end
    for i=1, #expired do finishTween(expired[i]) end
end

function tween.count()
  local c = 0
  for _,v in pairs(tweens) do
    c = c + 1
  end
  return c
end

-- Executes a function every 'period' seconds
-- 
-- Period is not accurate & is bound to frame
-- rate. The callback will be called a maximum
-- of once per frame.
--
-- If for instance your period is equivalent
-- to 1.5x framerate, when the period elapses
-- mid-frame the callback will not be executed
-- until the following frame.
function tween.interval(period, callback)
    local loop = tween.delay(period)
    loop.callback = function()
        
        -- Call the interval callback
        callback(loop)
        
        -- Did the callback call stop?
        if loop.forceStop then
            return
        end
        
        -- Execute as accurately as we can
        -- so account for any overshoot of timing
        -- in the current invocation
        loop.time = math.max(0, period - loop.overshoot)
            
        -- Restart our tween
        tween.reset(loop)
        tween.play(loop)
    end
    
    function loop:stop()
        tween.stop(self)
        self.forceStop = true
    end
    
    return loop
end

end
------------------------------
-- GlobalWrappers.lua
------------------------------
do
local gmt = getmetatable(_G)
if gmt and gmt.___CODEA_PLUS then
    return
end
assert(gmt == nil, "Global metatable modifications are not compatible with 'Codea+'")
    
    
local wrappers = {}
    
gmt = {
    __index = function(t, k)
        local w = wrappers[k]
        if w then
            return function(...)
                w.wfn(w.fn, ...)
            end
        end
    end,
    __newindex = function(t, k, v)
        if wrappers[k] then
            wrappers[k].fn = v
        else
            rawset(t, k, v)
        end
    end,
    ___CODEA_PLUS = true
}
setmetatable(_G, gmt)

function __wrapGlobal(name, fn)
    wrappers[name] = {
        wfn = fn,
        fn = nil
    }
end

function __getWrappedGlobal(name)
    return wrappers[name].fn
end

end
------------------------------
-- Thread.lua
------------------------------
do
local socket = require('socket')

-- Thread management values
local threads = {}
local current_thread = nil
local sbFlushImage = image(1,1)
local async_queue = {}

-- Thread management task
local managementTask = tween.interval(0.001, function()
    
    -- Reset per frame flags
    for _,thread in ipairs(threads) do
        thread.done_this_frame = false
    end
    
    -- Keep running while threads are still executing
    local thread_ran = true
    while thread_ran do
        
        -- Reset flag
        thread_ran = false
        
        local deadThreads = {}
    
        for i,thread in ipairs(threads) do
            
            local dbg_thread = Debugger.thread
            
            -- Is the coroutine dead?
            if coroutine.status(thread.co) == "dead" then
                table.insert(deadThreads, i)
                
            -- Has the thread finished for this frame?
            elseif thread.done_this_frame then
                
            -- Ignore this thread if it's sleeping
            elseif thread.sleep_end and socket.gettime() < thread.sleep_end then
                
                -- If the sleeping thread is 'blocking' pretend we ran
                if thread.blocking then
                    thread_ran = true
                end
                
            -- If we're debugging, only run the debugged thread
            elseif dbg_thread ~= nil and thread ~= dbg_thread then
                
            -- All checks have passed. Resume the thread!
            else
                -- Track the current thread
                current_thread = thread
                thread_ran = true
                
                local runMainRet = {}
                while true do
                    local res = table.pack(coroutine.resume(thread.co, table.unpack(runMainRet)))
                    if not res[1] then
                        -- Propagate the error
                        error(string.format("\n%s\n\n%s\n\nIGNORE TRACE BELOW:", res[2], debug.traceback(thread.co)))
                    elseif res[2] then
                        -- Call the value returned from the resume
                        -- and resume again.
                        -- This is how a thread forces something to
                        -- run on the main thread.
                        current_thread = nil
                        runMainRet = table.pack(res[2](select(3, table.unpack(res))))
                        current_thread = thread
                    else
                        -- Pause execution
                        break
                    end
                end
                
                current_thread = nil
            end
        end
        
        -- Remove dead threads
        for i=#deadThreads,1,-1 do
            table.remove(threads, deadThreads[i])
        end
    end
        
    -- Render an offscreen sprite to flush the
    -- spritebatcher.
    -- This ensures sprite rendering done on
    -- threads is actually visible on screen.
    -- A wild guess is that the batcher is only
    -- flushed after the draw() call, and not
    -- after tween handling.
    spriteMode(CORNER)
    sprite(sbFlushImage, -1, -1)
end)
managementTask.noAsyncCallback = true
managementTask.ignoreDebug = true


--- @class Thread
-- Cooperative thread object based on Lua coroutines.
Thread = class()
Thread.__name = "Codea+ Thread"

--- @function Thread (func)
-- Thread constructor.
-- @param func Function to execute on the new thread.
-- @usage function setup()
--     -- Start a thread that will print its value once per frame
--     Thread(function()
--
--         -- Our thread local counter
--         local i = 0
--
--         -- Loop forever
--         while true do
--
--             -- Print and increment counter
--             print("Thread counter", i)
--             i = i + 1
--             
--             -- Yield the thread to stop execution for the current frame
--             yield()
--         end
--     end)
-- end
function Thread:init(fn, name)
    self.co = coroutine.create(fn)
    self.blocking = false
    self.name = name
    
    table.insert(threads, self)
end

--- @function Thread.runOnMain (func, ...)
-- Immediately executes the given function on the main thread.
--
-- Various Codea functions do not function correctly when used from a coroutine so use this to temporarily hand execution back to the main thread.
-- @param func Function to run on the main thread.
-- @param ... Parameters to supply to the provided function.
-- @usage -- Takes a screenshot from the main thread.
-- Thread.runOnMain(function()
--     screenshot = viewer.snapshot()
-- end)
-- @usage -- Prints a message from the main thread.
-- Thread.runOnMain(print, "Hello main thread!")
function Thread.runOnMain(func, ...)
    if coroutine.isyieldable() then
        return coroutine.yield(func, ...)
    end
    return func(...)
end

function Thread.runAsync(func, ...)
    if func == nil or Debugger.thread then return end
    local params = {...}
    for i,p in ipairs(params) do
        -- Convert userdata param to a table
        if type(p) == "userdata" and p.___getters then
            local np = {}
            for k,fn in pairs(p.___getters) do
                np[k] = fn(p)
            end
            params[i] = np
        end
    end
    table.insert(async_queue, { 
        fn = func,
        params = params
    })
end

--- @function Thread.callback (func, ignoreDebugger)
-- Wraps a function to ensure it is executed on a thread.
--
-- By default wrapped functions will not execute (or be queued) if the debugger is active. Set ignoreDebugger to true if the callback should be executed regardless of the debugger.
-- @param func The function to be wrapped.
-- @param ignoreDebugger If true then the function will execute regardless of the state of the debugger.
-- @return The wrapped function.
-- @usage local function success(data, statusCode, headers)
--     print(statusCode)
-- end
--
-- -- If this weren't wrapped, we couldn't call
-- -- dbg() from inside the callback
-- local failure = Thread.callback(function(err)
--     print(err)
--     dbg()
-- end)
--
-- Thread.runOnMain(
--     http.request,
--     "https://codeawebrepo.co.uk/manifest.json",
--     success, failure
-- )
function Thread.callback(fn, ignoreDebugger)
    return function(...)
        if fn == nil or (Debugger.thread and not ignoreDebugger) then return end
        table.insert(async_queue, { 
            fn = fn,
            params = { ... }
        })
    end
end

--- @function Thread.wrapMain (func)
-- Wraps a function to ensure it is executed on the main thread.
-- @param func The function to wrap.
-- @return The wrapped function
function Thread.wrapMain(fn)
    return function(...)
        return Thread.runOnMain(fn, ...)
    end
end

--- @function Thread.current ()
-- Get the current thread.
-- @return The currently running thread, or nil if the main thread is running.
function Thread.current()
    return current_thread
end

--- @function Thread.allThreads ()
-- Get all active threads.
-- @return A table containing all active threads.
function Thread.allThreads()
    return threads
end

--- @function Thread.sleep (duration)
-- Put the current thread to sleep for a period of time.
-- @param duration Duration of sleep in seconds
function Thread.sleep(duration)
    current_thread.sleep_end = socket.gettime() + duration
    yield()
end

--- @function yield (fast)
-- Yield execution on the current thread.
-- This allows cooperative multitasking during a frame & across frames.
-- @param fast (default false)
-- When false, execution will not resume on this thread until the next frame at the earliest.
-- When true, execution can return to this thread in the current frame.
function yield(fast)
    if not coroutine.isyieldable() then
        error("Unable to yield!")
    end
    current_thread.done_this_frame = (fast == nil or fast == false)
    coroutine.yield()
end

--- @function abort ()
-- Halts all threads and stops execution
function abort()
    objc.warning("abort.")
    threads = {}
    yield()
end

__wrapGlobal("touched", Thread.runAsync)
__wrapGlobal("hover", Thread.runAsync)
__wrapGlobal("scroll", Thread.runAsync)
__wrapGlobal("pinch", Thread.runAsync)
__wrapGlobal("keyboard", Thread.runAsync)
-- __wrapGlobal("willClose", Thread.runAsync) -- Don't wrap this

-- Async callback thread
Thread(function()
    
    -- We're a blocking thread as callbacks should
    -- be handled in the same frame as they were issued
    current_thread.blocking = true
    
    while true do
        for _,cb in ipairs(async_queue) do
            cb.fn(table.unpack(cb.params))
        end
        async_queue = {}
        
        yield()
    end
    
end)

-- These should do nothing
__wrapGlobal("setup", function() end)
__wrapGlobal("draw", function() end)

-- Thread to call setup() & draw()
Thread(function()
    
    -- We're a blocking thread
    current_thread.blocking = true
    
    -- Call setup() once
    local fn = __getWrappedGlobal("setup")
    if fn then fn() end
    
    -- Call draw() repeatedly
    while true do
        local fn = __getWrappedGlobal("draw")
        if fn then fn() end
        yield()
    end
end)
end
------------------------------
-- Semaphore.lua
------------------------------
do
--- @class Semaphore
-- Semaphore object for thread coordination
Semaphore = class()

--- @function Semaphore ()
-- Semaphore constructor.
function Semaphore:init()
    self.count = 0
    self.waiting = 0
end

--- @function Semaphore:signal ()
-- Increments the semaphore value.
function Semaphore:signal()
    self.count = self.count + (count or 1)
end

--- @function Semaphore:signalAll ()
-- Increase the semaphore value by the number of threads currently waiting on this semaphore.
function Semaphore:signalAll()
    self.count = self.count + self.waiting
end

--- @function Semaphore:wait (fast)
-- Yield the current thread until the semaphore value is > 0.
-- This will return immediately if the semaphore value is already > 0.
-- @param fast (default false)
-- When false, execution will not resume on this thread until the next frame at the earliest.
-- When true, execution can return to this thread in the current frame.
function Semaphore:wait(fast)
    self.waiting = self.waiting + 1
    while self.count == 0 do
        yield(fast)
    end
    self.count = self.count - 1
    self.waiting = self.waiting - 1
end

end
------------------------------
-- Async.lua
------------------------------
do
--- @class Promise
-- JavaScript like promises
Promise = class()
local Chain = class()

Promise.__name = "Promise"
--- @function Promise (func)
-- Promise constructor.
-- @param func Function to execute on a new thread. This function should accept 2 parameters, resolve & reject.
-- @return New Promise object
function Promise:init(func)
    
    -- No results
    self.val = nil
    self.err = nil
    
    self.sem = Semaphore()
    
    -- Chained functions
    self.chains = {}
    
    -- Run the provided function on a separate thread
    if func then
        Thread(function()
            func(function(...)
                self:resolve(...)
            end,
            function(err)
                self:reject(err)
            end)
        end)
    end
end

--- @function Promise:resolve (...)
-- Resolve the promise with the provided values.
-- @param ... values to resolve the promise with. These will be passed to any chained promises.
function Promise:resolve(...)
    self.val = {...}
        
    -- Pass along promise chain
    for _,chain in ipairs(self.chains) do
        chain:fulfill(self.val)
    end
        
    -- Resume threads that are waiting on us
    self.sem:signalAll()
end

--- @function Promise:reject (err)
-- Reject the promise with the provided error.
-- @param err Error to reject the promise with. This will be passed to any chained promises.
function Promise:reject(err)
    self.err = err
    
    -- Pass along promise chain
    for _,chain in ipairs(self.chains) do
        chain:reject(err)
    end
    
    -- Resume threads that are waiting on us
    self.sem:signalAll()
end

--- @function Promise:next (onFullfilled, onRejected)
-- Chain handler functions to be called when the promise is resolved.
-- The next() method returns a Promise. It takes up to two arguments: callback functions for the success and failure cases of the Promise.
-- @param onFullfilled (optional) A Function called if the Promise is fulfilled. This function can have multiple arguments, the fulfillment values. If it is not a function, it is internally replaced with an "Identity" function (it returns the received argument).
-- @param onRejected (optional) A Function called if the Promise is rejected. This function has one argument, the rejection reason. If it is not a function, it is internally replaced with a "Thrower" function (it throws an error it received as argument).
-- @return A new promise to be resolved with the result of onFullfilled or onRejected.
function Promise:next(onFulfilled, onRejected)
    local chain = Chain(onFulfilled, onRejected)
    table.insert(self.chains, chain)
    
    if self.val then
        chain:fulfill(self.val)
    elseif self.err then
        chain:reject(self.err)
    end
    
    return chain.promise
end

--- @function Promise:catch (onRejected)
function Promise:catch(onRejected)
    return self:next(nil, onRejected)
end

--- @function Promise:finally (onFinally)
function Promise:finally(onFinally)
    local fn = function()
        return onFinally()
    end
    return self:next(fn, fn)
end

--- @function Promise.all (promises)
function Promise.all(promises)
    local pr = Promise()
    
    local count = #promises
    local res = {}
    
    local function onRejected(err)
        pr:reject(err)
    end
    
    for i,p in ipairs(promises) do
        p:next(function(...)
            -- If not rejected already
            if pr.err == nil then
                res[i] = {...} -- add the result
                
                -- Resolve the 'all' promise if
                -- all promises have resolved.
                count = count - 1
                if count == 0 then
                    pr:resolve(res)
                end
            end
        end, onRejected)
    end
    
    return pr
end

--- @function Promise:await (fast)
function Promise:await(fast)
    if self.val == nil and self.err == nil then
        self.sem:wait(fast)
    end
    if self.err then error(self.err:gsub("^.-:[0-9]+: ", ""), 2) end
    if self.val then return table.unpack(self.val) end
end

local function identity(arg)
    return arg
end

local function thrower(err)
    error(err:gsub("Async:[0-9]+: ", "")) -- Strip error line
end

function Chain:init(onFulfilled, onRejected)
    self.promise = Promise()
    self.onFullfilled = onFulfilled or identity
    self.onRejected = onRejected or thrower
end

function Chain:fulfill(val)
    Thread(function()
        local r = table.pack(pcall(self.onFullfilled, table.unpack(val)))
        if r[1] then
            self.promise:resolve(select(2, table.unpack(r)))
        else
            self.promise:reject(r[2])
        end
    end)
end

function Chain:reject(err)
    Thread(function()
        local r = table.pack(pcall(self.onRejected, err))
        if r[1] then
            self.promise:resolve(select(2, table.unpack(r)))
        else
            self.promise:reject(r[2])
        end
    end)
end

--- @function async (func)
-- Returns a wrapped function that when called will return a promise.
--
-- The called function will be executed asynchronously on a new thread and the returned promise will be resolved with the result of the function call.
--
-- This allows any function to be turned into an asynchronous function with little effort.
-- @param func The function to be wrapped.
-- @return The wrapped function
function async(func)
    return function(...)
        local params = {...}
        return Promise(function(resolve, reject)
            local res = table.pack(pcall(func, table.unpack(params)))
            if res[1] == true then
                -- Resolve & return all results
                resolve(select(2, table.unpack(res)))
            else
                -- Reject and return error message
                reject(res[2])
            end
        end)
    end
end

--- @function await (promise, fast)
-- Wait for the given promise to resolve.
-- @param promise The promise to wait for.
-- @param fast (default false)
-- When false, execution will not resume on this thread until the next frame at the earliest.
-- When true, execution can return to this thread in the current frame.
-- @return The values the promise was resolved with.
function await(promise, fast)
    return promise:await(fast)
end

end
------------------------------
-- ThreadCompat.lua
------------------------------
do
-- Some of Codea's builtin functions or objects
-- work incorrectly from within a coroutine.
-- The functions below hijack the metatable's of such
-- objects to ensure that they are only accessed from
-- the main thread (not a coroutine)

local wrapFunction = nil
local wrapTable = nil
local wrapUserdata = nil

local function wrap(lazy, ...)
    local res = {}
    
    for _,obj in ipairs({...}) do
        local t = type(obj)
        if obj == _G then -- Don't wrap the global table
            table.insert(res, obj)
        elseif t == "function" then
            table.insert(res, wrapFunction(obj))
        elseif t == "table" then
            table.insert(res, wrapTable(obj, lazy))
        elseif t == "userdata" then
            table.insert(res, wrapUserdata(obj))
        else
            table.insert(res, obj)
        end
    end
    
    return table.unpack(res)
end

wrapFunction = function(fn)
    return Thread.wrapMain(fn)
end

wrapTable = function(t, lazy)
    
    if lazy then
        
        -- Original metatable
        local omt = getmetatable(t)
        
        -- If we've already been converted, return table
        if omt and omt.___mainAccess then return t end
        
        -- Create a shallow clone of the table
        -- and clear the original to force it into
        -- our new metatable below
        local clone = {}
        for k, v in pairs(t) do
            clone[k] = v
            rawset(t, k, nil)
        end
        for k, v in ipairs(t) do
            clone[k] = v
            rawset(t, k, nil)
        end
        
        -- If the original metatable IS the table,
        -- the clone should do the same.
        if omt == t then
            setmetatable(clone, clone)
            omt = clone -- Consider the clone the new original mt
        else
            setmetatable(clone, omt)
        end
        
        -- Our new metatable
        local mt = {
            ___lazyObj = clone,
            ___mainAccess = true,
            
            __index = Thread.wrapMain(function(t, k)
                return wrap(true, clone[k])
            end),
            
            __newindex = Thread.wrapMain(function(t, k, v)
                clone[k] = v
            end),
            
            
        }
        
        setmetatable(t, mt)
    else
        
        -- Get our metatable
        local mt = getmetatable(t)
        
        -- Get our lazy object
        local lazyObj = mt.___lazyObj
        
        -- Copy all the lazy object values back
        -- into our table (wrapped this time)
        for k, v in pairs(lazyObj) do
            rawset(t, k, wrap(true, v))
        end
        for k, v in ipairs(lazyObj) do
            rawset(t, k, wrap(true, v))
        end
        
        -- We don't need our lazy object any more
        mt.___lazyObj = nil
    end
    
    return t
end

wrapUserdata = function(ud)
    local mt = getmetatable(ud)
            
    -- No metatable? Just add the userdata obj
    -- or already converted?
    if mt == nil or rawget(mt, "___mainAccess") == true then return ud end
    
    -- Set the 'converted' flag            
    rawset(mt, "___mainAccess", true)
    
    -- Custom __index function
    local __index = mt.__index
    mt.__index = function(t, k)
        -- '__cache' gets special non-wrapped treatment
        -- due to some Codea internals likely accessing
        -- its values directly using rawget.
        if k == "__cache" then
            return __index(t, k)
        end
        
        return wrap(true, Thread.runOnMain(__index, t, k))
    end
    
    -- Wraps the named metamethod in order to force it
    -- to execute on the main thread.
    local function wrapmm(name)
        local mm = rawget(mt, name)
        if type(mm) == "function" then
            -- Force it to execute on the main thread
            -- & ensure return values are also wrapped
            rawset(mt, name, function(...)
                return wrap(true, Thread.runOnMain(mm, ...))
            end)
        elseif mm then
            -- Ensure the value is adequately thread safe
            rawset(mt, name, wrap(true, mm))
        end
    end
    
    wrapmm("__add")
    wrapmm("__sub")
    wrapmm("__mul")
    wrapmm("__div")
    wrapmm("__mod")
    wrapmm("__pow")
    wrapmm("__unm")
    wrapmm("__idiv")
    wrapmm("__band")
    wrapmm("__bor")
    wrapmm("__bxor")
    wrapmm("__bnot")
    wrapmm("__shl")
    wrapmm("__shr")
    wrapmm("__add")
    wrapmm("__concat")
    wrapmm("__len")
    wrapmm("__eq")
    wrapmm("__lt")
    wrapmm("__le")
    wrapmm("__newindex")
    wrapmm("__call")
    wrapmm("__name")
    wrapmm("__close")
                
    return ud
end

-- Helper function
function accessOnMain(obj)
    return wrap(true, obj)
end

-- Override pairs & ipairs to ensure objects
-- with access forced to the main thread are
-- fully populated before iterating
local _pairs = pairs
function pairs(t)
    local mt = getmetatable(t)
    if mt and rawget(mt, "___lazyObj") then
        wrap(false, t)
    end
    return _pairs(t)
end

local _ipairs = ipairs
function ipairs(t)
    local mt = getmetatable(t)
    if mt and rawget(mt, "___lazyObj") then
        wrap(false, t)
    end
    return _ipairs(t)
end

-- Thread compatibility
function useThreadCompatLayer()
    layout = accessOnMain(layout)
    viewer = accessOnMain(viewer)
    asset = accessOnMain(asset)
    readText = accessOnMain(readText)
    saveText = accessOnMain(saveText)
    craft = accessOnMain(craft)
end

-- Use the compat layer by default
useThreadCompatLayer()

end
------------------------------
-- Utils.lua
------------------------------
do
-- Various functions overridden to run on the main thread.
-- If this is not done, most of these functions will simply
-- not work at all from a Thread.

--- @module http
-- Modifications to Codea's built-in http.request function.

--- @function http.request (url, success, fail, parameters)
-- Identical to the built-in function but callbacks will be executed on a thread to allow for debugging within the callbacks.
-- @param url URL to request
-- @param success Callback to be executed in the case of a successful request.
-- @param fail Callback to be executed in the case of a failed request.
-- @param parameters Request parameters.
local http_request = http.request
http.request = function(url, success, fail, parameters)
    Thread.runOnMain(http_request, url, Thread.callback(success, true), Thread.callback(fail, true), parameters)
end

--- @function http.requestProm (url, parameters)
-- Identical to the built-in function but returns results using a Promise object.
-- @param url URL to request
-- @param parameters Request parameters.
-- @return Promise object that will be resolved or rejected with the results of the request.
function http.requestProm(url, parameters)
    return Promise(function(resolve, reject)
        http.request(url, resolve, reject, parameters)
    end)
end

--- @function http.requestSync (url, parameters)
-- Identical to the built-in function but blocks waiting until the request completes (success or failure).
-- @param url URL to request
-- @param parameters Request parameters.
-- @return In the case of a successful request, a table in the form { response, statusCode, headers }
-- In the case of a failed request, a table in the form { nil, errorMessage }
function http.requestSync(url, parameters)
    
    local result = nil
    local sem = Semaphore()
    
    Thread.runOnMain(http_request, url,
        function(response, status, headers) -- Success
            result = { response, status, headers }
            sem:signal()
        end,
        function(err) -- Fail
            result = { nil, err }
            sem:signal()
        end, parameters)
    
    sem:wait()
    
    return table.unpack(result)
end

--- @module Util
-- Various useful functions & utilities.

--- @function registerTerminator (func)
-- Registers a function to be called when the project is stopped.
-- This is akin to Codea's 'willClose()' function but allows multiple callback functions to be registered.
-- @param func The function to be registered as a terminator.

--- @function addTerminator (func)
-- Alias of registerTerminator().

--- @function unregisterTerminator (func)
-- Removes a previously registered function from the list of terminator functions.
-- @param func The function to be unregistered.

--- @function removeTerminator (func)
-- Alias of unregisterTerminator().
local terminators = {}
function registerTerminator(func)
    terminators[func] = func
end
function unregisterTerminator(func)
    terminators[func] = nil
end
addTerminator = registerTerminator
removeTerminator = unregisterTerminator
__wrapGlobal("willClose", function(fn)
    for _,term in pairs(terminators) do
        term() -- call registered terminator
    end
    fn() -- call original
end)
-- blank willClose to initialise the wrapper above
function willClose() end

--- @function enableLogging (silent, logfile)
-- Enables logging all 'print' output to the project tab with the given name.
--
-- This also prefixes all output with the current time.
-- @param silent If true, nothing will be printed to the sidepanel, only the logfile.
-- @param logfile Name of the project tab to write our log to.
local _print = print
function enableLogging(silent, logfile)
    logfile = logfile or "output"
    saveProjectTab(logfile, "")
    local f = io.open(asset.path .. "/" .. logfile .. ".lua", "w")
    f:write("--[==========[\n]==========]")
    f:flush()
    print = function(...)
        local time = os.date("%X")
        if silent ~= true then _print(time, ...) end
        f:seek("cur", -12)
        f:write(time)
        for _,v in ipairs({...}) do
            f:write('\t' .. tostring(v))
        end
        f:write('\n]==========]')
        f:flush()
    end
end
    
end
------------------------------
-- Debugger.lua
------------------------------
do
--- @module Debugger
    
Debugger = {
    thread = nil,
    debug_threads = false,
}

local saved_values = {}
local saved_viewer_mode = nil

local info = {
    traceback = {},
    locals = {},
    upvalues = {}
}

local function gatherDebugInfo()
    
    local traceback = info.traceback
    local locals = info.locals
    local upvalues = info.upvalues
    
    local level = 1
    while true do
        local info = debug.getinfo(Debugger.thread.co, level + 2, "Sl")
        if not info then break end
        
        locals[level] = {}
        upvalues[level] = {}
        
        if info.what == "C" then   -- is a C function?
            table.insert(traceback, string.format("[%d] C function", level))
            
        else   -- a Lua function
            table.insert(traceback, string.format("[%d]%s:%d", level, info.short_src, info.currentline))
            
            -- Gather locals
            local i = 1
            while true do
                local name, val = debug.getlocal(Debugger.thread.co, level + 2, i)
                if not name then break end
                locals[level][name] = { val, i }
                i = i + 1
            end
            
            -- Gather upvalues
            local func = debug.getinfo(Debugger.thread.co, level + 2).func
            i = 1
            while true do
                local name, val = debug.getupvalue(func, i)
                if not name then break end
                upvalues[level][name] = { val, i }
                i = i + 1
            end
            
        end
        level = level + 1
    end
    
    info.traceback = table.concat(traceback, "\n")
end



local function continue()
    Debugger.thread = nil
    _G.c = saved_values.c
    _G.l = saved_values.l
    _G.sl = saved_values.sl
    _G.t = saved_values.t
    
    
    -- Clear traceback, locals & upvalues
    info.traceback = {}
    info.locals = {}
    info.upvalues = {}
end

-- Locals
local function locals(name, level)
    
    -- Default to display first stack frame
    name = name or 1
    
    if type(name) == "string" then
        if level == nil then
            
            for i=1,#info.locals do
                
                -- Check locals first
                local v = info.locals[i][name]
                if v == nil then
                    -- Then upvalues
                    v = info.upvalues[i][name]
                end
                
                if v then
                    v = v[1]
                    local typ = type(v)
                    if typ == "table" or typ == "function" then
                        print(name .. " = " .. tostring(v))
                    else
                        print(name .. " = " .. typ .. ": " .. tostring(v))
                    end
                    return v
                end
            end
            
            objc.warning("No local '" .. name .. "' in any stack frame")
        else
            if level > #info.locals then
                objc.warning("Invalid stack frame: " .. level)
                return
            end
            
            -- Check locals first
            local v = info.locals[level][name]
            if v == nil then
                -- Then upvalues
                v = info.upvalues[level][name]
            end
            
            if v then
                v = v[1]
                local typ = type(v)
                if typ == "table" or typ == "function" then
                    print(name .. " = " .. tostring(v))
                else
                    print(name .. " = " .. typ .. ": " .. tostring(v))
                end
                return v
            else
                objc.warning("No local '" .. name .. "' in stack frame " .. level)
            end
        end
            
    elseif type(name) == "number" then
        if name > #info.locals then
            objc.warning("Invalid stack frame: " .. name)
            return
        end
        
        local out = {"Locals (level " .. name .. "):"}
        for k,v in pairs(info.locals[name]) do
            
            v = v[1]
            local typ = type(v)
            
            if typ == "table" or typ == "function" then
                table.insert(out, k .. " = " .. tostring(v))
            else
                local vs = tostring(v)
                if string.len(vs) > 32 then
                    vs = "..."
                end
                
                table.insert(out, k .. " = " .. typ .. ": " .. vs)
            end
        end
        
        -- Append upvalues
        for k,v in pairs(info.upvalues[name]) do
            v = v[1]
            local typ = type(v)
            
            if typ == "table" or typ == "function" then
                table.insert(out, k .. " = " .. tostring(v))
            else
                local vs = tostring(v)
                if string.len(vs) > 32 then
                    vs = "..."
                end
                
                table.insert(out, k .. " = " .. typ .. ": " .. vs)
            end
        end
        
        print(table.concat(out, "\n"))
    else
        objc.warning("l: Invalid parameter type. Expects string or number.")
    end
end

local function setlocal(name, value, level)
    
    if level == nil then
        
        for i=1,#info.locals do
            -- Check locals first
            local v = info.locals[i][name]
            
            if v then
                -- Set the new value
                debug.setlocal(Debugger.thread.co, i + 2, v[2], value)
                v[1] = value
                print("Local '" .. name .. "' set in stack frame " .. i)
                return
            else
                -- Then upvalues
                v = info.upvalues[i][name]
                if v then
                    -- Set the new value
                    local func = debug.getinfo(Debugger.thread.co, i + 2).func
                    debug.setupvalue(func, v[2], value)
                    v[1] = value
                    print("Local '" .. name .. "' set in stack frame " .. i)
                    return
                end
            end
        end
            
        objc.warning("No local '" .. name .. "' in any stack frame")
    else
        if level > #info.locals then
            objc.warning("Invalid stack frame: " .. level)
            return
        end
        
        -- Check locals first
        local v = info.locals[level][name]
            
        if v then
            -- Set the new value
            debug.setlocal(Debugger.thread.co, level, v[2], value)
            v[1] = value
            print("Local '" .. name .. "' set in stack frame " .. level)
            return
        else
            -- Then upvalues
            v = info.upvalues[level][name]
            if v then
                -- Set the new value
                local func = debug.getinfo(Debugger.thread.co, level + 2).func
                debug.setupvalue(func, v[2], value)
                v[1] = value
                print("Local '" .. name .. "' set in stack frame " .. level)
                return
            end
        end
        
        objc.warning("No local '" .. name .. "' in stack frame " .. level)
    end
end

local function traceback()
    print("Trace:\n" .. info.traceback)
end

local function launch_debugger()
    Debugger.thread = Thread.current()
    saved_values.c = _G.c
    saved_values.l = _G.l
    saved_values.sl = _G.sl
    saved_values.t = _G.t
    _G.c = continue
    _G.l = locals
    _G.sl = setlocal
    _G.t = traceback
    
    -- Show sidepanel
    saved_viewer_mode = viewer.mode
    viewer.mode = OVERLAY
    
    -- Pause physics
    physics.pause()
end

--- @function dbg (condition)
-- Breaks into the debugger. If a condition is provided, the debugger is only triggered if the condition evaluates to true.
-- @param condition (optional) If true, the debugger will be triggered.
function dbg(condition)
    if condition == nil or condition == true then
        launch_debugger()
        gatherDebugInfo()
        
        -- Capture the current screen image
        local screen_image = viewer.snapshot()
        
        -- Print backtrace
        print("Debug break:\n" .. info.traceback)
        
        -- Yield until the debugger deactivates itself
        while Debugger.thread ~= nil do
            yield()
            
            -- Draw the current screen so we don't flicker
            pushStyle()
            spriteMode(CORNER)
            sprite(screen_image, 0, 0, WIDTH, HEIGHT)
            popStyle()
        end
        
        -- Restore sidepanel mode
        viewer.mode = saved_viewer_mode
    
        -- Resume physics
        physics.resume()
    end
end


--- @function c ()
-- Exit the debugger and continue execution.
--
-- <b>Important:</b> This is only to be called manually using the Codea sidepanel while the debugger is active.

--- @function l (nameOrLevel, level)
-- View local variables.
--
-- When called with no arguments this prints all local variables at the stack frame of the active 'dbg()' call.
--
-- Note: When listed, string values are omitted. Use l("&lt;variable_name&gt;", [stack_level]) to view the full value.
--
-- <b>Important:</b> This is only to be called manually using the Codea sidepanel while the debugger is active.
-- @param nameOrLevel (optional) Name of the variable to read <b>OR</b> stack level to list.
-- @param level (optional) Level to search for the given local variable name.
-- @usage -- Prints all local variables in the current stack frame
-- l()
-- @usage -- Prints all local variables in the next stack frame up (1 == current frame)
-- l(2)
-- @usage -- Prints the value of the local variable 'i'
-- -- in the first stack frame it is found.
-- l("i")
-- @usage -- Prints the value of the local variable 'i' in the 3rd stack frame.
-- l("i", 3)

--- @function sl (name, value, level)
-- Set a local variable.
--
-- <b>Important:</b> This is only to be called manually using the Codea sidepanel while the debugger is active.

--- @function t ()
-- Print the backtrace again.
--
-- <b>Important:</b> This is only to be called manually using the Codea sidepanel while the debugger is active.
 
end
------------------------------
-- AutoDoc.lua
------------------------------
do
local HTML_TEMPLATE = nil
local MD_TEMPLATE = nil

--- @module AutoDoc
-- Generates HTML documentation for the current project.
    
local function GenerateHTMLDoc()
    local blocks = {}
    
    for _,tab in ipairs(listProjectTabs()) do
        -- Ignore dependency files
        if tab:match("^_dep_") then goto continue end
        
        -- Read the source
        local f = io.open(asset.path .. "/" .. tab .. ".lua", "r")
        local src = f:read("*a") f:close()
        
        -- Identify doc comments
        for comment in src:gmatch("%-(%-%- .-)\n[^%-]") do
            
            local block = {}
            
            local i = 1
            local param, val
            while true do
                i, vi, param = comment:find("%-%- @([^%s]+) ?", i)
                if i == nil then break end -- No param
                
                -- Get the val
                _, _, val = comment:find(" ?(.-)\n%-%- ?@", vi+1)
                if val == nil then
                    val = comment:sub(vi+1)
                end
                
                -- Strip comment dashes
                if val then
                    val = val:gsub("\n%-%- ?", "\n")
                end
                
                i = i + 1
                if param == "param" then
                    block.params = block.params or {}
                    local name, desc = val:match("(.-)%s(.*)")
                    table.insert(block.params, {name, desc:gsub("\n", "<br>")})
                elseif param == "usage" then
                    block.examples = block.examples or {}
                    table.insert(block.examples, val)
                elseif param == "function" or param == "class" or param == "module" then
                    local sig = val:match("[^\n]*")
                    local desc = val:match("\n(.*)") or ""
                    block[param] = sig
                    block.desc = desc:gsub("\n", "<br>")
                else
                    block[param] = val:gsub("\n", "<br>")
                end
            end
            
            table.insert(blocks, block)
        end
        
        ::continue::
    end
    
    local html = HTML_TEMPLATE
    local name = (asset.name):gsub("%.codea", "")
    html = html:gsub("{{{TITLE}}}", name)
                                                
    -- Generate contents
    local contents = {}
    for _,block in ipairs(blocks) do
        if block.class then
            table.insert(contents, string.format([[<li><a href="#Class_%s">Class: %s</a></li>]], block.class, block.class))
        end
        
        if block.module then
            table.insert(contents, string.format([[<li><a href="#Module_%s">Module: %s</a></li>]], block.module, block.module))
        end
    end
    html = html:gsub("{{{CONTENTS}}}", table.concat(contents))
                                        
    -- Generate class function list
    local cfl = {}
    for _,block in ipairs(blocks) do
        if block.class then
            -- End previous class function list
            if #cfl > 0 then
                table.insert(cfl, "</table>")
            end
            table.insert(cfl, string.format([[<h2><a href="#Class_%s">Class: %s</a></h2>]], block.class, block.class))
            table.insert(cfl, [[<table class="function_list">]])
            
        elseif block.module then
            -- End previous class function list
            if #cfl > 0 then
                table.insert(cfl, "</table>")
            end
            table.insert(cfl, string.format([[<h2><a href="#Module_%s">Module: %s</a></h2>]], block.module, block.module))
            table.insert(cfl, [[<table class="function_list">]])
        end
        
        if block["function"] then
            table.insert(cfl, string.format([[
<tr>
<td class="name" nowrap><a href="#%s">%s</a></td>
<td class="summary">%s</td>
</tr>
            ]], block["function"], block["function"], block.desc:match("(.-%.)") or ""))
        end
    end
    if #cfl > 0 then
        table.insert(cfl, "</table>")
    end
    html = html:gsub("{{{CLASS_FUNCTION_LIST}}}", table.concat(cfl))
    
    -- Generate sections
    local sections = {}
    local curSection = nil
    for _,block in ipairs(blocks) do
        if block.class then
            -- End previous section
            if curSection then
                table.insert(curSection,"</dl>")
                table.insert(sections, table.concat(curSection))
            end

            curSection = {}
            table.insert(curSection, string.format([[
<h2 class="section-header has-description"><a name="Class_%s"></a>Class: %s</h2>
<div class="section-description">
%s
</div>
<dl class="function">
            ]], block.class, block.class, block.desc))

        elseif block.module then
            -- End previous section
            if curSection then
                table.insert(curSection,"</dl>")
                table.insert(sections, table.concat(curSection))
            end

            curSection = {}
            table.insert(curSection, string.format([[
<h2 class="section-header has-description"><a name="Module_%s"></a>Module: %s</h2>
<div class="section-description">
%s
</div>
<dl class="function">
            ]], block.module, block.module, block.desc or "No description."))
        end
        
        if block["function"] then
            if not curSection then
                objc.warning("No section!")
            end
            
            -- Add function heading
            local heading = string.format([[
            <dt>
            <a name = "%s"></a>
            <strong>%s</strong>
            </dt>
            <dd>
            %s
            ]], block["function"], block["function"], block.desc or "No description.")
            table.insert(curSection, heading)
            
            -- Add function parameters
            if block.params then
                local params = {
                    [[<h3>Parameters:</h3>
                    <ul>]]
                }
                for k, v in pairs(block.params) do
                    local s = string.format([[<li><span class="parameter">%s</span>
                    %s</li>]], v[1], v[2])
                    table.insert(params, s)
                end
                table.insert(params, "</ul>")
                table.insert(curSection, table.concat(params))
            end
            
            -- Add function returns
            if block["return"] then
                table.insert(curSection, string.format("<h3>Returns:</h3><ol>%s</ol>", block["return"]))
            end
            
            -- Add usage examples
            if block.examples then
                local examples = {
                    "<h3>Usage:</h3><ul>"
                }
                
                local function keywordSub(kw, code)
                    return code:gsub('^(' .. kw .. ')', '<span class="keyword">%1</span>')
                end
                
                for _, code in ipairs(block.examples) do
                    code = string.format('<li><pre><code class="language-lua">%s</code></pre></li>', code)
                    table.insert(examples, code)
                end
                
                table.insert(examples, "</ul>")
                table.insert(curSection, table.concat(examples))
            end
            
            -- Close the description block
            table.insert(curSection, "</dd>")
        end
    end
    if curSection then
        table.insert(sections, table.concat(curSection))
    end
    html = html:gsub("{{{SECTIONS}}}", table.concat(sections))
    
    -- Add the generated time
    html = html:gsub("{{{DATE_TIME}}}", os.date())
    
    return html
end

local function GenerateMDDoc()
    local blocks = {}
    
    for _,tab in ipairs(listProjectTabs()) do
        -- Ignore dependency files
        if tab:match("^_dep_") then goto continue end
        
        -- Read the source
        local f = io.open(asset.path .. "/" .. tab .. ".lua", "r")
        local src = f:read("*a") f:close()
        
        -- Identify doc comments
        for comment in src:gmatch("%-(%-%- .-)\n[^%-]") do
            
            local block = {}
            
            local i = 1
            local param, val
            while true do
                i, vi, param = comment:find("%-%- @([^%s]+) ?", i)
                if i == nil then break end -- No param
                
                -- Get the val
                _, _, val = comment:find(" ?(.-)\n%-%- ?@", vi+1)
                if val == nil then
                    val = comment:sub(vi+1)
                end
                
                -- Strip comment dashes
                if val then
                    val = val:gsub("\n%-%- ?", "\n")
                end
                
                i = i + 1
                if param == "param" then
                    block.params = block.params or {}
                    local name, desc = val:match("(.-)%s(.*)")
                    table.insert(block.params, {name, desc:gsub("\n", "<br>")})
                elseif param == "usage" then
                    block.examples = block.examples or {}
                    table.insert(block.examples, val)
                elseif param == "function" or param == "class" or param == "module" then
                    local sig = val:match("[^\n]*")
                    local desc = val:match("\n(.*)") or ""
                    block[param] = sig
                    block.desc = desc:gsub("\n", "<br>")
                else
                    block[param] = val:gsub("\n", "<br>")
                end
            end
            
            table.insert(blocks, block)
        end
        
        ::continue::
    end
    
    local html = MD_TEMPLATE
    html = html:gsub("{{{TITLE}}}", "Codea+")
                                        
    -- Generate class function list
    local cfl = {}
    for _,block in ipairs(blocks) do
        if block.class then
            table.insert(cfl, string.format("## **Class:** %s\n", block.class))
            
        elseif block.module then
            table.insert(cfl, string.format("## **Module:** %s\n", block.module))
        end
        
        if block["function"] then
            table.insert(cfl, string.format("`%s`: %s\n", block["function"], block.desc:match("(.-%.)") or ""))
        end
    end
    html = html:gsub("{{{CLASS_FUNCTION_LIST}}}", table.concat(cfl))
    
    -- Generate sections
    local sections = {}
    local curSection = nil
    for _,block in ipairs(blocks) do
        if block.class then
            if curSection then table.insert(sections, table.concat(curSection)) end
            curSection = {}
            table.insert(curSection, string.format("<br>\n<br>\n<br>\n# **Class:** %s\n### %s\n", block.class, block.desc))

        elseif block.module then
            if curSection then table.insert(sections, table.concat(curSection)) end
            curSection = {}
            table.insert(curSection, string.format("<br>\n<br>\n<br>\n# **Module:** %s\n### %s\n", block.module, block.desc))
        end
        
        if block["function"] then
            if not curSection then
                objc.warning("No section!")
            end
            
            -- Add function heading
            local heading = string.format("\n\n---\n\n## **%s**\n%s\n", block["function"], block.desc or "No description.")
            table.insert(curSection, heading)
            
            -- Add function parameters
            if block.params then
                local params = {
                    "### **Parameters:**\n"
                }
                for k, v in pairs(block.params) do
                    table.insert(params, string.format("- `%s`: %s\n", v[1], v[2]))
                end
                table.insert(curSection, table.concat(params))
            end
            
            -- Add function returns
            if block["return"] then
                table.insert(curSection, string.format("### **Returns:** %s\n", block["return"]))
            end
            
            -- Add usage examples
            --[[
            if block.examples then
                local examples = {
                    "<h3>Usage:</h3><ul>"
                }
                
                local function keywordSub(kw, code)
                    return code:gsub('^(' .. kw .. ')', '<span class="keyword">%1</span>')
                end
                
                for _, code in ipairs(block.examples) do
                    code = string.format('<li><pre><code class="language-lua">%s</code></pre></li>', code)
                    table.insert(examples, code)
                end
                
                table.insert(examples, "</ul>")
                table.insert(curSection, table.concat(examples))
            end
            ]]
        end
    end
    html = html:gsub("{{{SECTIONS}}}", table.concat(sections))
    
    -- Add the generated time
    html = html:gsub("{{{DATE_TIME}}}", os.date())
    
    return html
end

--- @function AutoDoc (display)
-- Generates HTML documentation from the current project source and optionally displays it in a Fullscreen webview.
-- @param display (optional) Display a fullscreen webview and do not return to the caller.
-- @return HTML string
-- @usage -- Return the generated html to the caller
-- local html = AutoDoc()
-- @usage AutoDoc(true) -- This never returns 
function AutoDoc(display)
    
    local html = GenerateHTMLDoc()
    
    -- Save docs
    local f = io.open(asset.path .. "/docs.html", "w")
    f:write(html)
    f:close()
    
    -- Return the html to the caller
    if not display then
        return html
    end
    
    -- No sidebar
    viewer.mode = FULLSCREEN
    
    -- Open the DovViewer
    local dv = DocViewer(asset.docs)
    
    -- Do not return
    while true do
        yield()
    end
end

function AutoDocMD()
    return GenerateMDDoc()
end

MD_TEMPLATE = [==[
# {{{TITLE}}}

---

{{{CLASS_FUNCTION_LIST}}}
{{{SECTIONS}}}

---

Generated by Codea+ AutoDoc
Last updated {{{DATE_TIME}}}
]==]

HTML_TEMPLATE = [==[
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
   "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8"/>
<head>
    <title>{{{TITLE}}}</title>
    <style>
        /*
        Steppers, 2022
        This stylesheet was taken from LDoc.
        I claim no ownership over this content.
        */
        
        /* BEGIN RESET
        Copyright (c) 2010, Yahoo! Inc. All rights reserved.
        Code licensed under the BSD License:
        http://developer.yahoo.com/yui/license.html
        version: 2.8.2r1
        */
        html {
            color: #000;
            background: #FFF;
        }
        body,div,dl,dt,dd,ul,ol,li,h1,h2,h3,h4,h5,h6,pre,code,form,fieldset,legend,input,button,textarea,p,blockquote,th,td {
            margin: 0;
            padding: 0;
        }
        table {
            border-collapse: collapse;
            border-spacing: 0;
        }
        fieldset,img {
            border: 0;
        }
        address,caption,cite,code,dfn,em,strong,th,var,optgroup {
            font-style: inherit;
            font-weight: inherit;
        }
        del,ins {
            text-decoration: none;
        }
        li {
            margin-left: 20px;
        }
        caption,th {
            text-align: left;
        }
        h1,h2,h3,h4,h5,h6 {
            font-size: 100%;
            font-weight: bold;
        }
        q:before,q:after {
            content: '';
        }
        abbr,acronym {
            border: 0;
            font-variant: normal;
        }
        sup {
            vertical-align: baseline;
        }
        sub {
            vertical-align: baseline;
        }
        legend {
            color: #000;
        }
        input,button,textarea,select,optgroup,option {
            font-family: inherit;
            font-size: inherit;
            font-style: inherit;
            font-weight: inherit;
        }
        input,button,textarea,select {*font-size:100%;
        }
        /* END RESET */
        
        body {
            margin-left: 1em;
            margin-right: 1em;
            font-family: arial, helvetica, geneva, sans-serif;
            background-color: #ffffff; margin: 0px;
        }
        
        code, tt { font-family: monospace; font-size: 1.1em; }
        span.parameter { font-family:monospace; }
        span.parameter:after { content:":"; }
        span.types:before { content:"("; }
        span.types:after { content:")"; }
        .type { font-weight: bold; font-style:italic }
        
        body, p, td, th { font-size: .95em; line-height: 1.2em;}
        
        p, ul { margin: 10px 0 0 0px;}
        
        strong { font-weight: bold;}
        
        em { font-style: italic;}
        
        h1 {
            font-size: 1.5em;
            margin: 20px 0 20px 0;
        }
        h2, h3, h4 { margin: 15px 0 10px 0; }
        h2 { font-size: 1.25em; }
        h3 { font-size: 1.15em; }
        h4 { font-size: 1.06em; }
        
        a:link { font-weight: bold; color: #004080; text-decoration: none; }
        a:visited { font-weight: bold; color: #006699; text-decoration: none; }
        a:link:hover { text-decoration: underline; }
        
        hr {
            color:#cccccc;
            background: #00007f;
            height: 1px;
        }
        
        blockquote { margin-left: 3em; }
        
        ul { list-style-type: disc; }
        
        p.name {
            font-family: "Andale Mono", monospace;
            padding-top: 1em;
        }
        
        pre {
            background-color: rgb(245, 245, 245);
            border: 1px solid #C0C0C0; /* silver */
            padding: 4px;
            margin: 10px 0 10px 0;
            overflow: auto;
            font-family: "Andale Mono", monospace;
        }
        
        pre.example {
            font-size: .85em;
        }
        
        table.index { border: 1px #00007f; }
        table.index td { text-align: left; vertical-align: top; }
        
        #container {
            margin-left: 1em;
            margin-right: 1em;
            background-color: #f0f0f0;
        }
        
        #product {
            text-align: center;
            border-bottom: 1px solid #cccccc;
            background-color: #ffffff;
        }
        
        #product big {
            font-size: 2em;
        }
        
        #main {
            background-color: #f0f0f0;
            border-left: 2px solid #cccccc;
        }
        
        #navigation {
            float: left;
            width: 14em;
            vertical-align: top;
            background-color: #f0f0f0;
            overflow: visible;
        }
        
        #navigation h2 {
            background-color:#e7e7e7;
            font-size:1.1em;
            color:#000000;
            text-align: left;
            padding:0.2em;
            border-top:1px solid #dddddd;
            border-bottom:1px solid #dddddd;
        }
        
        #navigation ul
        {
            font-size:1em;
            list-style-type: none;
            margin: 1px 1px 10px 1px;
        }
        
        #navigation li {
            text-indent: -1em;
            display: block;
            margin: 3px 0px 0px 22px;
        }
        
        #navigation li li a {
            margin: 0px 3px 0px -1em;
        }
        
        #content {
            margin-left: 14em;
            padding: 1em;
            width: 700px;
            border-left: 2px solid #cccccc;
            border-right: 2px solid #cccccc;
            background-color: #ffffff;
        }
        
        #about {
            clear: both;
            padding: 5px;
            border-top: 2px solid #cccccc;
            background-color: #ffffff;
        }
        
        @media print {
            body {
                font: 12pt "Times New Roman", "TimeNR", Times, serif;
            }
            a { font-weight: bold; color: #004080; text-decoration: underline; }
        
            #main {
                background-color: #ffffff;
                border-left: 0px;
            }
        
            #container {
                margin-left: 2%;
                margin-right: 2%;
                background-color: #ffffff;
            }
        
            #content {
                padding: 1em;
                background-color: #ffffff;
            }
        
            #navigation {
                display: none;
            }
            pre.example {
                font-family: "Andale Mono", monospace;
                font-size: 10pt;
                page-break-inside: avoid;
            }
        }
        
        table.module_list {
            border-width: 1px;
            border-style: solid;
            border-color: #cccccc;
            border-collapse: collapse;
        }
        table.module_list td {
            border-width: 1px;
            padding: 3px;
            border-style: solid;
            border-color: #cccccc;
        }
        table.module_list td.name { background-color: #f0f0f0; min-width: 200px; }
        table.module_list td.summary { width: 100%; }
        
        
        table.function_list {
            border-width: 1px;
            border-style: solid;
            border-color: #cccccc;
            border-collapse: collapse;
        }
        table.function_list td {
            border-width: 1px;
            padding: 3px;
            border-style: solid;
            border-color: #cccccc;
        }
        table.function_list td.name { background-color: #f0f0f0; min-width: 200px; }
        table.function_list td.summary { width: 100%; }
        
        ul.nowrap {
            overflow:auto;
            white-space:nowrap;
        }
        
        dl.table dt, dl.function dt {border-top: 1px solid #ccc; padding-top: 1em;}
        dl.table dd, dl.function dd {padding-bottom: 1em; margin: 10px 0 0 20px;}
        dl.table h3, dl.function h3 {font-size: .95em;}
        
        /* stop sublists from having initial vertical space */
        ul ul { margin-top: 0px; }
        ol ul { margin-top: 0px; }
        ol ol { margin-top: 0px; }
        ul ol { margin-top: 0px; }
        
        /* make the target distinct; helps when we're navigating to a function */
        a:target + * {
        	background-color: #FF9;
        }
    </style>
    <style>
        pre code.hljs{display:block;overflow-x:auto;padding:1em}code.hljs{padding:3px 5px}.hljs{background:#fff;color:#000}.hljs-comment,.hljs-quote,.hljs-variable{color:green}.hljs-built_in,.hljs-keyword,.hljs-name,.hljs-selector-tag,.hljs-tag{color:#00f}.hljs-addition,.hljs-attribute,.hljs-literal,.hljs-section,.hljs-string,.hljs-template-tag,.hljs-template-variable,.hljs-title,.hljs-type{color:#a31515}.hljs-deletion,.hljs-meta,.hljs-selector-attr,.hljs-selector-pseudo{color:#2b91af}.hljs-doctag{color:grey}.hljs-attr{color:red}.hljs-bullet,.hljs-link,.hljs-symbol{color:#00b0e8}.hljs-emphasis{font-style:italic}.hljs-strong{font-weight:700}
    </style>
    <script>
/*!
  Highlight.js v11.5.1 (git: b8f233c8e2)
  (c) 2006-2022 Ivan Sagalaev and other contributors
  License: BSD-3-Clause
 */
var hljs=function(){"use strict";var e={exports:{}};function t(e){
return e instanceof Map?e.clear=e.delete=e.set=()=>{
throw Error("map is read-only")}:e instanceof Set&&(e.add=e.clear=e.delete=()=>{
throw Error("set is read-only")
}),Object.freeze(e),Object.getOwnPropertyNames(e).forEach((n=>{var i=e[n]
;"object"!=typeof i||Object.isFrozen(i)||t(i)})),e}
e.exports=t,e.exports.default=t;var n=e.exports;class i{constructor(e){
void 0===e.data&&(e.data={}),this.data=e.data,this.isMatchIgnored=!1}
ignoreMatch(){this.isMatchIgnored=!0}}function r(e){
return e.replace(/&/g,"&amp;").replace(/</g,"&lt;").replace(/>/g,"&gt;").replace(/"/g,"&quot;").replace(/'/g,"&#x27;")
}function s(e,...t){const n=Object.create(null);for(const t in e)n[t]=e[t]
;return t.forEach((e=>{for(const t in e)n[t]=e[t]})),n}const o=e=>!!e.kind
;class a{constructor(e,t){
this.buffer="",this.classPrefix=t.classPrefix,e.walk(this)}addText(e){
this.buffer+=r(e)}openNode(e){if(!o(e))return;let t=e.kind
;t=e.sublanguage?"language-"+t:((e,{prefix:t})=>{if(e.includes(".")){
const n=e.split(".")
;return[`${t}${n.shift()}`,...n.map(((e,t)=>`${e}${"_".repeat(t+1)}`))].join(" ")
}return`${t}${e}`})(t,{prefix:this.classPrefix}),this.span(t)}closeNode(e){
o(e)&&(this.buffer+="</span>")}value(){return this.buffer}span(e){
this.buffer+=`<span class="${e}">`}}class c{constructor(){this.rootNode={
children:[]},this.stack=[this.rootNode]}get top(){
return this.stack[this.stack.length-1]}get root(){return this.rootNode}add(e){
this.top.children.push(e)}openNode(e){const t={kind:e,children:[]}
;this.add(t),this.stack.push(t)}closeNode(){
if(this.stack.length>1)return this.stack.pop()}closeAllNodes(){
for(;this.closeNode(););}toJSON(){return JSON.stringify(this.rootNode,null,4)}
walk(e){return this.constructor._walk(e,this.rootNode)}static _walk(e,t){
return"string"==typeof t?e.addText(t):t.children&&(e.openNode(t),
t.children.forEach((t=>this._walk(e,t))),e.closeNode(t)),e}static _collapse(e){
"string"!=typeof e&&e.children&&(e.children.every((e=>"string"==typeof e))?e.children=[e.children.join("")]:e.children.forEach((e=>{
c._collapse(e)})))}}class l extends c{constructor(e){super(),this.options=e}
addKeyword(e,t){""!==e&&(this.openNode(t),this.addText(e),this.closeNode())}
addText(e){""!==e&&this.add(e)}addSublanguage(e,t){const n=e.root
;n.kind=t,n.sublanguage=!0,this.add(n)}toHTML(){
return new a(this,this.options).value()}finalize(){return!0}}function g(e){
return e?"string"==typeof e?e:e.source:null}function d(e){return f("(?=",e,")")}
function u(e){return f("(?:",e,")*")}function h(e){return f("(?:",e,")?")}
function f(...e){return e.map((e=>g(e))).join("")}function p(...e){const t=(e=>{
const t=e[e.length-1]
;return"object"==typeof t&&t.constructor===Object?(e.splice(e.length-1,1),t):{}
})(e);return"("+(t.capture?"":"?:")+e.map((e=>g(e))).join("|")+")"}
function b(e){return RegExp(e.toString()+"|").exec("").length-1}
const m=/\[(?:[^\\\]]|\\.)*\]|\(\??|\\([1-9][0-9]*)|\\./
;function E(e,{joinWith:t}){let n=0;return e.map((e=>{n+=1;const t=n
;let i=g(e),r="";for(;i.length>0;){const e=m.exec(i);if(!e){r+=i;break}
r+=i.substring(0,e.index),
i=i.substring(e.index+e[0].length),"\\"===e[0][0]&&e[1]?r+="\\"+(Number(e[1])+t):(r+=e[0],
"("===e[0]&&n++)}return r})).map((e=>`(${e})`)).join(t)}
const x="[a-zA-Z]\\w*",w="[a-zA-Z_]\\w*",y="\\b\\d+(\\.\\d+)?",_="(-?)(\\b0[xX][a-fA-F0-9]+|(\\b\\d+(\\.\\d*)?|\\.\\d+)([eE][-+]?\\d+)?)",k="\\b(0b[01]+)",v={
begin:"\\\\[\\s\\S]",relevance:0},O={scope:"string",begin:"'",end:"'",
illegal:"\\n",contains:[v]},N={scope:"string",begin:'"',end:'"',illegal:"\\n",
contains:[v]},M=(e,t,n={})=>{const i=s({scope:"comment",begin:e,end:t,
contains:[]},n);i.contains.push({scope:"doctag",
begin:"[ ]*(?=(TODO|FIXME|NOTE|BUG|OPTIMIZE|HACK|XXX):)",
end:/(TODO|FIXME|NOTE|BUG|OPTIMIZE|HACK|XXX):/,excludeBegin:!0,relevance:0})
;const r=p("I","a","is","so","us","to","at","if","in","it","on",/[A-Za-z]+['](d|ve|re|ll|t|s|n)/,/[A-Za-z]+[-][a-z]+/,/[A-Za-z][a-z]{2,}/)
;return i.contains.push({begin:f(/[ ]+/,"(",r,/[.]?[:]?([.][ ]|[ ])/,"){3}")}),i
},S=M("//","$"),R=M("/\\*","\\*/"),j=M("#","$");var A=Object.freeze({
__proto__:null,MATCH_NOTHING_RE:/\b\B/,IDENT_RE:x,UNDERSCORE_IDENT_RE:w,
NUMBER_RE:y,C_NUMBER_RE:_,BINARY_NUMBER_RE:k,
RE_STARTERS_RE:"!|!=|!==|%|%=|&|&&|&=|\\*|\\*=|\\+|\\+=|,|-|-=|/=|/|:|;|<<|<<=|<=|<|===|==|=|>>>=|>>=|>=|>>>|>>|>|\\?|\\[|\\{|\\(|\\^|\\^=|\\||\\|=|\\|\\||~",
SHEBANG:(e={})=>{const t=/^#![ ]*\//
;return e.binary&&(e.begin=f(t,/.*\b/,e.binary,/\b.*/)),s({scope:"meta",begin:t,
end:/$/,relevance:0,"on:begin":(e,t)=>{0!==e.index&&t.ignoreMatch()}},e)},
BACKSLASH_ESCAPE:v,APOS_STRING_MODE:O,QUOTE_STRING_MODE:N,PHRASAL_WORDS_MODE:{
begin:/\b(a|an|the|are|I'm|isn't|don't|doesn't|won't|but|just|should|pretty|simply|enough|gonna|going|wtf|so|such|will|you|your|they|like|more)\b/
},COMMENT:M,C_LINE_COMMENT_MODE:S,C_BLOCK_COMMENT_MODE:R,HASH_COMMENT_MODE:j,
NUMBER_MODE:{scope:"number",begin:y,relevance:0},C_NUMBER_MODE:{scope:"number",
begin:_,relevance:0},BINARY_NUMBER_MODE:{scope:"number",begin:k,relevance:0},
REGEXP_MODE:{begin:/(?=\/[^/\n]*\/)/,contains:[{scope:"regexp",begin:/\//,
end:/\/[gimuy]*/,illegal:/\n/,contains:[v,{begin:/\[/,end:/\]/,relevance:0,
contains:[v]}]}]},TITLE_MODE:{scope:"title",begin:x,relevance:0},
UNDERSCORE_TITLE_MODE:{scope:"title",begin:w,relevance:0},METHOD_GUARD:{
begin:"\\.\\s*[a-zA-Z_]\\w*",relevance:0},END_SAME_AS_BEGIN:e=>Object.assign(e,{
"on:begin":(e,t)=>{t.data._beginMatch=e[1]},"on:end":(e,t)=>{
t.data._beginMatch!==e[1]&&t.ignoreMatch()}})});function I(e,t){
"."===e.input[e.index-1]&&t.ignoreMatch()}function T(e,t){
void 0!==e.className&&(e.scope=e.className,delete e.className)}function L(e,t){
t&&e.beginKeywords&&(e.begin="\\b("+e.beginKeywords.split(" ").join("|")+")(?!\\.)(?=\\b|\\s)",
e.__beforeBegin=I,e.keywords=e.keywords||e.beginKeywords,delete e.beginKeywords,
void 0===e.relevance&&(e.relevance=0))}function B(e,t){
Array.isArray(e.illegal)&&(e.illegal=p(...e.illegal))}function D(e,t){
if(e.match){
if(e.begin||e.end)throw Error("begin & end are not supported with match")
;e.begin=e.match,delete e.match}}function H(e,t){
void 0===e.relevance&&(e.relevance=1)}const P=(e,t)=>{if(!e.beforeMatch)return
;if(e.starts)throw Error("beforeMatch cannot be used with starts")
;const n=Object.assign({},e);Object.keys(e).forEach((t=>{delete e[t]
})),e.keywords=n.keywords,e.begin=f(n.beforeMatch,d(n.begin)),e.starts={
relevance:0,contains:[Object.assign(n,{endsParent:!0})]
},e.relevance=0,delete n.beforeMatch
},C=["of","and","for","in","not","or","if","then","parent","list","value"]
;function $(e,t,n="keyword"){const i=Object.create(null)
;return"string"==typeof e?r(n,e.split(" ")):Array.isArray(e)?r(n,e):Object.keys(e).forEach((n=>{
Object.assign(i,$(e[n],t,n))})),i;function r(e,n){
t&&(n=n.map((e=>e.toLowerCase()))),n.forEach((t=>{const n=t.split("|")
;i[n[0]]=[e,U(n[0],n[1])]}))}}function U(e,t){
return t?Number(t):(e=>C.includes(e.toLowerCase()))(e)?0:1}const z={},K=e=>{
console.error(e)},W=(e,...t)=>{console.log("WARN: "+e,...t)},X=(e,t)=>{
z[`${e}/${t}`]||(console.log(`Deprecated as of ${e}. ${t}`),z[`${e}/${t}`]=!0)
},G=Error();function Z(e,t,{key:n}){let i=0;const r=e[n],s={},o={}
;for(let e=1;e<=t.length;e++)o[e+i]=r[e],s[e+i]=!0,i+=b(t[e-1])
;e[n]=o,e[n]._emit=s,e[n]._multi=!0}function F(e){(e=>{
e.scope&&"object"==typeof e.scope&&null!==e.scope&&(e.beginScope=e.scope,
delete e.scope)})(e),"string"==typeof e.beginScope&&(e.beginScope={
_wrap:e.beginScope}),"string"==typeof e.endScope&&(e.endScope={_wrap:e.endScope
}),(e=>{if(Array.isArray(e.begin)){
if(e.skip||e.excludeBegin||e.returnBegin)throw K("skip, excludeBegin, returnBegin not compatible with beginScope: {}"),
G
;if("object"!=typeof e.beginScope||null===e.beginScope)throw K("beginScope must be object"),
G;Z(e,e.begin,{key:"beginScope"}),e.begin=E(e.begin,{joinWith:""})}})(e),(e=>{
if(Array.isArray(e.end)){
if(e.skip||e.excludeEnd||e.returnEnd)throw K("skip, excludeEnd, returnEnd not compatible with endScope: {}"),
G
;if("object"!=typeof e.endScope||null===e.endScope)throw K("endScope must be object"),
G;Z(e,e.end,{key:"endScope"}),e.end=E(e.end,{joinWith:""})}})(e)}function V(e){
function t(t,n){
return RegExp(g(t),"m"+(e.case_insensitive?"i":"")+(e.unicodeRegex?"u":"")+(n?"g":""))
}class n{constructor(){
this.matchIndexes={},this.regexes=[],this.matchAt=1,this.position=0}
addRule(e,t){
t.position=this.position++,this.matchIndexes[this.matchAt]=t,this.regexes.push([t,e]),
this.matchAt+=b(e)+1}compile(){0===this.regexes.length&&(this.exec=()=>null)
;const e=this.regexes.map((e=>e[1]));this.matcherRe=t(E(e,{joinWith:"|"
}),!0),this.lastIndex=0}exec(e){this.matcherRe.lastIndex=this.lastIndex
;const t=this.matcherRe.exec(e);if(!t)return null
;const n=t.findIndex(((e,t)=>t>0&&void 0!==e)),i=this.matchIndexes[n]
;return t.splice(0,n),Object.assign(t,i)}}class i{constructor(){
this.rules=[],this.multiRegexes=[],
this.count=0,this.lastIndex=0,this.regexIndex=0}getMatcher(e){
if(this.multiRegexes[e])return this.multiRegexes[e];const t=new n
;return this.rules.slice(e).forEach((([e,n])=>t.addRule(e,n))),
t.compile(),this.multiRegexes[e]=t,t}resumingScanAtSamePosition(){
return 0!==this.regexIndex}considerAll(){this.regexIndex=0}addRule(e,t){
this.rules.push([e,t]),"begin"===t.type&&this.count++}exec(e){
const t=this.getMatcher(this.regexIndex);t.lastIndex=this.lastIndex
;let n=t.exec(e)
;if(this.resumingScanAtSamePosition())if(n&&n.index===this.lastIndex);else{
const t=this.getMatcher(0);t.lastIndex=this.lastIndex+1,n=t.exec(e)}
return n&&(this.regexIndex+=n.position+1,
this.regexIndex===this.count&&this.considerAll()),n}}
if(e.compilerExtensions||(e.compilerExtensions=[]),
e.contains&&e.contains.includes("self"))throw Error("ERR: contains `self` is not supported at the top-level of a language.  See documentation.")
;return e.classNameAliases=s(e.classNameAliases||{}),function n(r,o){const a=r
;if(r.isCompiled)return a
;[T,D,F,P].forEach((e=>e(r,o))),e.compilerExtensions.forEach((e=>e(r,o))),
r.__beforeBegin=null,[L,B,H].forEach((e=>e(r,o))),r.isCompiled=!0;let c=null
;return"object"==typeof r.keywords&&r.keywords.$pattern&&(r.keywords=Object.assign({},r.keywords),
c=r.keywords.$pattern,
delete r.keywords.$pattern),c=c||/\w+/,r.keywords&&(r.keywords=$(r.keywords,e.case_insensitive)),
a.keywordPatternRe=t(c,!0),
o&&(r.begin||(r.begin=/\B|\b/),a.beginRe=t(a.begin),r.end||r.endsWithParent||(r.end=/\B|\b/),
r.end&&(a.endRe=t(a.end)),
a.terminatorEnd=g(a.end)||"",r.endsWithParent&&o.terminatorEnd&&(a.terminatorEnd+=(r.end?"|":"")+o.terminatorEnd)),
r.illegal&&(a.illegalRe=t(r.illegal)),
r.contains||(r.contains=[]),r.contains=[].concat(...r.contains.map((e=>(e=>(e.variants&&!e.cachedVariants&&(e.cachedVariants=e.variants.map((t=>s(e,{
variants:null},t)))),e.cachedVariants?e.cachedVariants:q(e)?s(e,{
starts:e.starts?s(e.starts):null
}):Object.isFrozen(e)?s(e):e))("self"===e?r:e)))),r.contains.forEach((e=>{n(e,a)
})),r.starts&&n(r.starts,o),a.matcher=(e=>{const t=new i
;return e.contains.forEach((e=>t.addRule(e.begin,{rule:e,type:"begin"
}))),e.terminatorEnd&&t.addRule(e.terminatorEnd,{type:"end"
}),e.illegal&&t.addRule(e.illegal,{type:"illegal"}),t})(a),a}(e)}function q(e){
return!!e&&(e.endsWithParent||q(e.starts))}class J extends Error{
constructor(e,t){super(e),this.name="HTMLInjectionError",this.html=t}}
const Y=r,Q=s,ee=Symbol("nomatch");var te=(e=>{
const t=Object.create(null),r=Object.create(null),s=[];let o=!0
;const a="Could not find the language '{}', did you forget to load/include a language module?",c={
disableAutodetect:!0,name:"Plain text",contains:[]};let g={
ignoreUnescapedHTML:!1,throwUnescapedHTML:!1,noHighlightRe:/^(no-?highlight)$/i,
languageDetectRe:/\blang(?:uage)?-([\w-]+)\b/i,classPrefix:"hljs-",
cssSelector:"pre code",languages:null,__emitter:l};function b(e){
return g.noHighlightRe.test(e)}function m(e,t,n){let i="",r=""
;"object"==typeof t?(i=e,
n=t.ignoreIllegals,r=t.language):(X("10.7.0","highlight(lang, code, ...args) has been deprecated."),
X("10.7.0","Please use highlight(code, options) instead.\nhttps://github.com/highlightjs/highlight.js/issues/2277"),
r=e,i=t),void 0===n&&(n=!0);const s={code:i,language:r};N("before:highlight",s)
;const o=s.result?s.result:E(s.language,s.code,n)
;return o.code=s.code,N("after:highlight",o),o}function E(e,n,r,s){
const c=Object.create(null);function l(){if(!O.keywords)return void M.addText(S)
;let e=0;O.keywordPatternRe.lastIndex=0;let t=O.keywordPatternRe.exec(S),n=""
;for(;t;){n+=S.substring(e,t.index)
;const r=y.case_insensitive?t[0].toLowerCase():t[0],s=(i=r,O.keywords[i]);if(s){
const[e,i]=s
;if(M.addText(n),n="",c[r]=(c[r]||0)+1,c[r]<=7&&(R+=i),e.startsWith("_"))n+=t[0];else{
const n=y.classNameAliases[e]||e;M.addKeyword(t[0],n)}}else n+=t[0]
;e=O.keywordPatternRe.lastIndex,t=O.keywordPatternRe.exec(S)}var i
;n+=S.substr(e),M.addText(n)}function d(){null!=O.subLanguage?(()=>{
if(""===S)return;let e=null;if("string"==typeof O.subLanguage){
if(!t[O.subLanguage])return void M.addText(S)
;e=E(O.subLanguage,S,!0,N[O.subLanguage]),N[O.subLanguage]=e._top
}else e=x(S,O.subLanguage.length?O.subLanguage:null)
;O.relevance>0&&(R+=e.relevance),M.addSublanguage(e._emitter,e.language)
})():l(),S=""}function u(e,t){let n=1;const i=t.length-1;for(;n<=i;){
if(!e._emit[n]){n++;continue}const i=y.classNameAliases[e[n]]||e[n],r=t[n]
;i?M.addKeyword(r,i):(S=r,l(),S=""),n++}}function h(e,t){
return e.scope&&"string"==typeof e.scope&&M.openNode(y.classNameAliases[e.scope]||e.scope),
e.beginScope&&(e.beginScope._wrap?(M.addKeyword(S,y.classNameAliases[e.beginScope._wrap]||e.beginScope._wrap),
S=""):e.beginScope._multi&&(u(e.beginScope,t),S="")),O=Object.create(e,{parent:{
value:O}}),O}function f(e,t,n){let r=((e,t)=>{const n=e&&e.exec(t)
;return n&&0===n.index})(e.endRe,n);if(r){if(e["on:end"]){const n=new i(e)
;e["on:end"](t,n),n.isMatchIgnored&&(r=!1)}if(r){
for(;e.endsParent&&e.parent;)e=e.parent;return e}}
if(e.endsWithParent)return f(e.parent,t,n)}function p(e){
return 0===O.matcher.regexIndex?(S+=e[0],1):(I=!0,0)}function b(e){
const t=e[0],i=n.substr(e.index),r=f(O,e,i);if(!r)return ee;const s=O
;O.endScope&&O.endScope._wrap?(d(),
M.addKeyword(t,O.endScope._wrap)):O.endScope&&O.endScope._multi?(d(),
u(O.endScope,e)):s.skip?S+=t:(s.returnEnd||s.excludeEnd||(S+=t),
d(),s.excludeEnd&&(S=t));do{
O.scope&&M.closeNode(),O.skip||O.subLanguage||(R+=O.relevance),O=O.parent
}while(O!==r.parent);return r.starts&&h(r.starts,e),s.returnEnd?0:t.length}
let m={};function w(t,s){const a=s&&s[0];if(S+=t,null==a)return d(),0
;if("begin"===m.type&&"end"===s.type&&m.index===s.index&&""===a){
if(S+=n.slice(s.index,s.index+1),!o){const t=Error(`0 width match regex (${e})`)
;throw t.languageName=e,t.badRule=m.rule,t}return 1}
if(m=s,"begin"===s.type)return(e=>{
const t=e[0],n=e.rule,r=new i(n),s=[n.__beforeBegin,n["on:begin"]]
;for(const n of s)if(n&&(n(e,r),r.isMatchIgnored))return p(t)
;return n.skip?S+=t:(n.excludeBegin&&(S+=t),
d(),n.returnBegin||n.excludeBegin||(S=t)),h(n,e),n.returnBegin?0:t.length})(s)
;if("illegal"===s.type&&!r){
const e=Error('Illegal lexeme "'+a+'" for mode "'+(O.scope||"<unnamed>")+'"')
;throw e.mode=O,e}if("end"===s.type){const e=b(s);if(e!==ee)return e}
if("illegal"===s.type&&""===a)return 1
;if(A>1e5&&A>3*s.index)throw Error("potential infinite loop, way more iterations than matches")
;return S+=a,a.length}const y=k(e)
;if(!y)throw K(a.replace("{}",e)),Error('Unknown language: "'+e+'"')
;const _=V(y);let v="",O=s||_;const N={},M=new g.__emitter(g);(()=>{const e=[]
;for(let t=O;t!==y;t=t.parent)t.scope&&e.unshift(t.scope)
;e.forEach((e=>M.openNode(e)))})();let S="",R=0,j=0,A=0,I=!1;try{
for(O.matcher.considerAll();;){
A++,I?I=!1:O.matcher.considerAll(),O.matcher.lastIndex=j
;const e=O.matcher.exec(n);if(!e)break;const t=w(n.substring(j,e.index),e)
;j=e.index+t}return w(n.substr(j)),M.closeAllNodes(),M.finalize(),v=M.toHTML(),{
language:e,value:v,relevance:R,illegal:!1,_emitter:M,_top:O}}catch(t){
if(t.message&&t.message.includes("Illegal"))return{language:e,value:Y(n),
illegal:!0,relevance:0,_illegalBy:{message:t.message,index:j,
context:n.slice(j-100,j+100),mode:t.mode,resultSoFar:v},_emitter:M};if(o)return{
language:e,value:Y(n),illegal:!1,relevance:0,errorRaised:t,_emitter:M,_top:O}
;throw t}}function x(e,n){n=n||g.languages||Object.keys(t);const i=(e=>{
const t={value:Y(e),illegal:!1,relevance:0,_top:c,_emitter:new g.__emitter(g)}
;return t._emitter.addText(e),t})(e),r=n.filter(k).filter(O).map((t=>E(t,e,!1)))
;r.unshift(i);const s=r.sort(((e,t)=>{
if(e.relevance!==t.relevance)return t.relevance-e.relevance
;if(e.language&&t.language){if(k(e.language).supersetOf===t.language)return 1
;if(k(t.language).supersetOf===e.language)return-1}return 0})),[o,a]=s,l=o
;return l.secondBest=a,l}function w(e){let t=null;const n=(e=>{
let t=e.className+" ";t+=e.parentNode?e.parentNode.className:""
;const n=g.languageDetectRe.exec(t);if(n){const t=k(n[1])
;return t||(W(a.replace("{}",n[1])),
W("Falling back to no-highlight mode for this block.",e)),t?n[1]:"no-highlight"}
return t.split(/\s+/).find((e=>b(e)||k(e)))})(e);if(b(n))return
;if(N("before:highlightElement",{el:e,language:n
}),e.children.length>0&&(g.ignoreUnescapedHTML||(console.warn("One of your code blocks includes unescaped HTML. This is a potentially serious security risk."),
console.warn("https://github.com/highlightjs/highlight.js/wiki/security"),
console.warn("The element with unescaped HTML:"),
console.warn(e)),g.throwUnescapedHTML))throw new J("One of your code blocks includes unescaped HTML.",e.innerHTML)
;t=e;const i=t.textContent,s=n?m(i,{language:n,ignoreIllegals:!0}):x(i)
;e.innerHTML=s.value,((e,t,n)=>{const i=t&&r[t]||n
;e.classList.add("hljs"),e.classList.add("language-"+i)
})(e,n,s.language),e.result={language:s.language,re:s.relevance,
relevance:s.relevance},s.secondBest&&(e.secondBest={
language:s.secondBest.language,relevance:s.secondBest.relevance
}),N("after:highlightElement",{el:e,result:s,text:i})}let y=!1;function _(){
"loading"!==document.readyState?document.querySelectorAll(g.cssSelector).forEach(w):y=!0
}function k(e){return e=(e||"").toLowerCase(),t[e]||t[r[e]]}
function v(e,{languageName:t}){"string"==typeof e&&(e=[e]),e.forEach((e=>{
r[e.toLowerCase()]=t}))}function O(e){const t=k(e)
;return t&&!t.disableAutodetect}function N(e,t){const n=e;s.forEach((e=>{
e[n]&&e[n](t)}))}
"undefined"!=typeof window&&window.addEventListener&&window.addEventListener("DOMContentLoaded",(()=>{
y&&_()}),!1),Object.assign(e,{highlight:m,highlightAuto:x,highlightAll:_,
highlightElement:w,
highlightBlock:e=>(X("10.7.0","highlightBlock will be removed entirely in v12.0"),
X("10.7.0","Please use highlightElement now."),w(e)),configure:e=>{g=Q(g,e)},
initHighlighting:()=>{
_(),X("10.6.0","initHighlighting() deprecated.  Use highlightAll() now.")},
initHighlightingOnLoad:()=>{
_(),X("10.6.0","initHighlightingOnLoad() deprecated.  Use highlightAll() now.")
},registerLanguage:(n,i)=>{let r=null;try{r=i(e)}catch(e){
if(K("Language definition for '{}' could not be registered.".replace("{}",n)),
!o)throw e;K(e),r=c}
r.name||(r.name=n),t[n]=r,r.rawDefinition=i.bind(null,e),r.aliases&&v(r.aliases,{
languageName:n})},unregisterLanguage:e=>{delete t[e]
;for(const t of Object.keys(r))r[t]===e&&delete r[t]},
listLanguages:()=>Object.keys(t),getLanguage:k,registerAliases:v,
autoDetection:O,inherit:Q,addPlugin:e=>{(e=>{
e["before:highlightBlock"]&&!e["before:highlightElement"]&&(e["before:highlightElement"]=t=>{
e["before:highlightBlock"](Object.assign({block:t.el},t))
}),e["after:highlightBlock"]&&!e["after:highlightElement"]&&(e["after:highlightElement"]=t=>{
e["after:highlightBlock"](Object.assign({block:t.el},t))})})(e),s.push(e)}
}),e.debugMode=()=>{o=!1},e.safeMode=()=>{o=!0
},e.versionString="11.5.1",e.regex={concat:f,lookahead:d,either:p,optional:h,
anyNumberOfTimes:u};for(const e in A)"object"==typeof A[e]&&n(A[e])
;return Object.assign(e,A),e})({});return te}()
;"object"==typeof exports&&"undefined"!=typeof module&&(module.exports=hljs);/*! `lua` grammar compiled for Highlight.js 11.5.1 */
(()=>{var e=(()=>{"use strict";return e=>{const t="\\[=*\\[",a="\\]=*\\]",n={
begin:t,end:a,contains:["self"]
},o=[e.COMMENT("--(?!\\[=*\\[)","$"),e.COMMENT("--\\[=*\\[",a,{contains:[n],
relevance:10})];return{name:"Lua",keywords:{$pattern:e.UNDERSCORE_IDENT_RE,
literal:"true false nil",
keyword:"and break do else elseif end for goto if in local not or repeat return then until while",
built_in:"_G _ENV _VERSION __index __newindex __mode __call __metatable __tostring __len __gc __add __sub __mul __div __mod __pow __concat __unm __eq __lt __le assert collectgarbage dofile error getfenv getmetatable ipairs load loadfile loadstring module next pairs pcall print rawequal rawget rawset require select setfenv setmetatable tonumber tostring type unpack xpcall arg self coroutine resume yield status wrap create running debug getupvalue debug sethook getmetatable gethook setmetatable setlocal traceback setfenv getinfo setupvalue getlocal getregistry getfenv io lines write close flush open output type read stderr stdin input stdout popen tmpfile math log max acos huge ldexp pi cos tanh pow deg tan cosh sinh random randomseed frexp ceil floor rad abs sqrt modf asin min mod fmod log10 atan2 exp sin atan os exit setlocale date getenv difftime remove time clock tmpname rename execute package preload loadlib loaded loaders cpath config path seeall string sub upper len gfind rep find match char dump gmatch reverse byte format gsub lower table setn insert getn foreachi maxn foreach concat sort remove"
},contains:o.concat([{className:"function",beginKeywords:"function",end:"\\)",
contains:[e.inherit(e.TITLE_MODE,{
begin:"([_a-zA-Z]\\w*\\.)*([_a-zA-Z]\\w*:)?[_a-zA-Z]\\w*"}),{className:"params",
begin:"\\(",endsWithParent:!0,contains:o}].concat(o)
},e.C_NUMBER_MODE,e.APOS_STRING_MODE,e.QUOTE_STRING_MODE,{className:"string",
begin:t,end:a,contains:[n],relevance:5}])}}})();hljs.registerLanguage("lua",e)
})(); hljs.highlightAll();
    </script>
</head>
<body>

<div id="container">

<div id="product">
	<div id="product_logo"></div>
	<div id="product_name"><big><b></b></big></div>
	<div id="product_description"></div>
</div> <!-- id="product" -->
<div id="main">
<!-- Menu -->
<div id="navigation">
<br/>
<h1>{{{TITLE}}}</h1>
<h2>Contents</h2>
<ul>
{{{CONTENTS}}}
</ul>
</div>
<div id="content">
<p></p>
{{{CLASS_FUNCTION_LIST}}}
<br/>
</dl>
{{{SECTIONS}}}
</dl>
</div> <!-- id="content" -->
</div> <!-- id="main" -->
<div id="about">
<i>Generated by Codea+ AutoDoc</a></i>
<i style="float:right;">Last updated {{{DATE_TIME}}}</i>
</div> <!-- id="about" -->
</div> <!-- id="container" -->
</body>
</html>
]==]

-- Generate docs
Thread(function()
    if AUTODOC_ENABLE == true then AutoDoc() end
end)

end
------------------------------
-- DocViewer.lua
------------------------------
do
-- Based on the 'JavaScript' lib available on WebRepo.
-- This is a highly stripped down version of the library
-- above intended to serve a static page only.

local WebView = class()
local WebViewURLSchemeHandler = class()

-- Alias
DocViewer = WebView

-- Default URL scheme handler 'codea://' reads resources from the
-- current project's asset folder.
function WebViewURLSchemeHandler:init()
    if WebViewURLSchemeHandler.Handler == nil then
        WebViewURLSchemeHandler.Handler = objc.delegate("WKURLSchemeHandler")
        
        local mimeType = {
            txt         = "text/plain",
            png         = "image/png",
            css         = "text/css",
            js          = "text/javascript"
        }
        
        WebViewURLSchemeHandler.Handler.webView_startURLSchemeTask_ = function(self, objWebView, objUrlSchemeTask)
                
            local url = objUrlSchemeTask.request.URL
                
            -- Get the file path
            local path = url.absoluteString:gsub("codea://", "")
            path = socket.url.unescape(path)
            local ext = path:match(".*%.(.-)$")
                
            -- Read the file into NSData
            local data = objc.cls.NSData:dataWithContentsOfFile_(asset.path .. "/" .. path)
                
            -- Create the response
            local response = objc.cls.NSHTTPURLResponse:alloc()
            response:initWithURL_statusCode_HTTPVersion_headerFields_(url, 200, "HTTP/1.1", {
                ["Access-Control-Allow-Origin"] = "*",
                ["Content-Type"] = mimeType[ext] or "application/octet-stream"
            })
                
            objUrlSchemeTask:didReceiveResponse_(response)
            objUrlSchemeTask:didReceiveData_(data)
            objUrlSchemeTask:didFinish()
        end
    end
    self.handler = WebViewURLSchemeHandler.Handler()
end

function WebViewURLSchemeHandler:register(config)
    config:setURLSchemeHandler_forURLScheme_(self.handler, "codea")
end

function WebView:init(htmlAsset)
    Thread.runOnMain(function()
        -- Create the WebView config
        local config = objc.cls.WKWebViewConfiguration()
        self.urlSchemeHandler = WebViewURLSchemeHandler()
        self.urlSchemeHandler:register(config)
        
        -- New view
        self.view = objc.cls.WKWebView:alloc()
        self.view:initWithFrame_configuration_(objc.viewer.view.bounds, config)
        
        -- Remove old WKWebView view
        if objc.viewer.view.subviews[2].Class == "WKWebView" then
            objc.viewer.view.subviews[2]:removeFromSuperview()
        end
        
        -- Place the WKWebView over the EAGLView
        objc.viewer.view:insertSubview_atIndex_(self.view, 1)
        
        -- Ensure cleanup is done
        registerTerminator(function()
            self.view:removeFromSuperview()
        end)
    end)
    
    -- Trigger a page load
    self:loadHTML(htmlAsset)
end

local function loadHTMLFromFile(view, htmlAsset)
    
    -- Escape spaces
    local path = htmlAsset.path:gsub(" ", "%%20")
    
    Thread.runOnMain(function()
        local request = objc.cls.NSURLRequest:requestWithURL_cachePolicy_timeoutInterval_(
            objc.cls.NSURL:URLWithString_("file://" .. path),
            objc.enum.NSURLRequestCachePolicy.NSURLRequestReloadIgnoringCacheData,
            5.0)
        view:loadRequest_(request)
    end)
end

function WebView:loadHTML(obj)
    loadHTMLFromFile(self.view, obj)
end

end
